from typing import Dict

import numpy as np

from labml import logger
from .. import Writer, Indicator
from ..indicators.artifacts import Artifact
from ..indicators.numeric import NumericIndicator
from labml.logger import Text


class ScreenWriter(Writer):
    def __init__(self):
        super().__init__()

        self._estimates = {}
        self._beta = 0.9
        self._beta_pow = {}
        self._last_printed_value = {}

    def update_estimate(self, k, v):
        if k not in self._estimates:
            self._estimates[k] = 0
            self._beta_pow[k] = 1.

        self._estimates[k] *= self._beta
        self._estimates[k] += (1 - self._beta) * v
        self._beta_pow[k] *= self._beta

    def get_empty_string(self, length, decimals):
        return ' ' * (length - 2 - decimals) + '-.' + '-' * decimals

    def get_value_string(self, k, v):
        if k not in self._estimates:
            assert v is None
            return self.get_empty_string(8, 2)

        estimate = self._estimates[k] / (1 - self._beta_pow[k])
        if abs(estimate) < 1e-9 or np.isnan(estimate):
            lg = 0
        else:
            lg = int(np.ceil(np.log10(abs(estimate)))) + 1

        decimals = 7 - lg
        decimals = max(1, decimals)
        decimals = min(6, decimals)

        fmt = "{v:8,." + str(decimals) + "f}"
        if v is None:
            return self.get_empty_string(8, decimals)
        else:
            return fmt.format(v=v)

    @staticmethod
    def __format_artifact(length: int, value: str):
        fmt = "{v:>" + str(length + 1) + "}"
        return fmt.format(v=value)

    def _get_indicator_string(self, indicators: Dict[str, Indicator]):
        parts = []

        for ind in indicators.values():
            if not isinstance(ind, NumericIndicator):
                continue
            if not ind.is_print:
                continue

            parts.append((f" {ind.name}: ", None))

            if not ind.is_empty():
                v = ind.get_mean()
                self.update_estimate(ind.name, v)
                value = self.get_value_string(ind.name, v)
                self._last_printed_value[ind.name] = value
                parts.append((value, Text.value))
            elif ind.name in self._last_printed_value:
                value = self._last_printed_value[ind.name]
                parts.append((value, Text.subtle))
            else:
                value = self.get_value_string(ind.name, None)
                parts.append((value, Text.subtle))

        return parts

    @staticmethod
    def _print_artifacts_list(table: Dict[str, int], artifacts: Dict[str, Artifact]):
        order = list(table.keys())
        if not len(order):
            return

        keys = {k for name in order for k in artifacts[name].keys()}
        for k in keys:
            for name in order:
                value = artifacts[name].get_string(k, artifacts)
                logger.log([(name, Text.key),
                            ": ",
                            (value, Text.value)])

    def _print_artifacts_table(self, table: Dict[str, int], artifacts: Dict[str, Artifact]):
        order = list(table.keys())
        if not len(order):
            return

        keys = []
        keys_set = set()

        for name in order:
            for k in artifacts[name].keys():
                if k not in keys_set:
                    keys_set.add(k)
                    keys.append(k)

        parts = [self.__format_artifact(table[name], name) for name in order]
        logger.log('|'.join(parts), Text.heading)

        for k in keys:
            parts = []
            for name in order:
                value = artifacts[name].get_string(k, artifacts)
                parts.append(self.__format_artifact(table[name], value))
            logger.log('|'.join(parts), Text.value)

    def _print_artifacts(self, indicators: Dict[str, Indicator]):
        table  = {}
        artifacts = {}
        for ind in indicators.values():
            if not isinstance(ind, Artifact):
                continue
            if not ind.is_print:
                continue
            if ind.is_empty():
                continue
            if not ind.is_indexed:
                ind.print_all()
                continue

            table[ind.name] = ind.get_print_length()
            artifacts[ind.name] = ind

        if sum(table.values()) > 100:
            self._print_artifacts_list(table, artifacts)
        else:
            self._print_artifacts_table(table, artifacts)

    def write(self, *,
              global_step: int,
              indicators: Dict[str, Indicator]):

        self._print_artifacts(indicators)

        return self._get_indicator_string(indicators)


import pathlib
import random
import string
from typing import Dict, Callable

import yaml

get_ipython: Callable


def yaml_load(s: str):
    return yaml.load(s, Loader=yaml.FullLoader)


def yaml_dump(obj: any):
    return yaml.dump(obj, default_flow_style=False)


def rm_tree(path_to_remove: pathlib.Path):
    if path_to_remove.is_dir():
        for f in path_to_remove.iterdir():
            if f.is_dir():
                rm_tree(f)
            else:
                f.unlink()
        path_to_remove.rmdir()
    else:
        path_to_remove.unlink()


def random_string(length=10):
    letters = string.ascii_lowercase
    return ''.join(random.choice(letters) for _ in range(length))


def is_ipynb():
    try:
        cfg = get_ipython().config
        if cfg['IPKernelApp'] is None:
            return False

        app: Dict = cfg['IPKernelApp']
        if len(app.keys()) > 0:
            return True
        else:
            return False
    except NameError:
        return False


def is_ipynb_pycharm():
    if not is_ipynb():
        return False

    if is_colab() or is_kaggle():
        return False

    import os
    if '_' not in os.environ:
        return True
    else:
        return False


def is_colab():
    import sys
    return 'google.colab' in sys.modules


def is_kaggle():
    import sys
    return 'kaggle_gcp' in sys.modules


if __name__ == '__main__':
    print(is_ipynb())



import matplotlib.pyplot as plt
import seaborn as sns

from labml.internal.analytics import cache as _cache
from labml.internal.analytics.indicators import IndicatorCollection
from labml.internal.analytics.matplotlib import MatPlotLibAnalytics as _MatPlotLibAnalytics

MATPLOTLIB = _MatPlotLibAnalytics()


def render_matplotlib(indicators: IndicatorCollection):
    r"""
    Creates a distribution plot distribution with MatPlotLib
    """
    _, ax = plt.subplots(figsize=(18, 9))
    for i, ind in enumerate(indicators):
        data = _cache.get_indicator_data(ind)
        MATPLOTLIB.render_density(ax, data, sns.color_palette()[i], ind.key)
    plt.show()



from typing import List, Union, Tuple, Callable, Optional

from IPython.core.display import display
from ipywidgets import HTML

from labml.internal.util.colors import StyleCode
from labml.internal.logger.destinations import Destination
from labml.internal.logger.destinations.ipynb import IpynbDestination
from labml.internal.util import is_kaggle

get_ipython: Callable


class IpynbPyCharmDestination(Destination):
    def __init__(self):
        self.__cell_lines = []
        self.__cell_count = 0
        self.html: Optional[HTML] = None

    def is_same_cell(self):
        try:
            cells = get_ipython().ev('len(In)')
        except NameError:
            return False
        if cells == self.__cell_count:
            return True

        self.__cell_count = cells
        self.__cell_lines = []

        return False

    def log(self, parts: List[Union[str, Tuple[str, Optional[StyleCode]]]], *,
            is_new_line: bool,
            is_reset: bool):
        tuple_parts = []
        for p in parts:
            if type(p) == str:
                text = p
                style = None
            else:
                text, style = p
            lines = text.split('\n')
            for line in lines[:-1]:
                tuple_parts.append((line, style))
                tuple_parts.append(('\n', None))
            tuple_parts.append((lines[-1], style))

        coded = [IpynbDestination.html_code(text, color) for text, color in tuple_parts]

        text = "".join(coded)
        lines = text.split('\n')

        if is_kaggle():
            attrs = 'style="color: #444; overflow-x: scroll;"'
        else:
            attrs = 'style="overflow-x: scroll;"'

        if self.is_same_cell():
            if coded:
                last = self.__cell_lines.pop()
                if is_reset:
                    self.__cell_lines += lines
                else:
                    self.__cell_lines += [last + lines[0]] + lines[1:]
            text = '\n'.join(self.__cell_lines)
            self.html.value = f"<pre {attrs}>{text}</pre>"
        else:
            self.__cell_lines = lines
            text = '\n'.join(self.__cell_lines)
            self.html = HTML(f"<pre  {attrs}>{text}</pre>")
            display(self.html)

        # print(len(self.__cell_lines), self.__cell_lines[-1], is_new_line)
        if is_new_line:
            self.__cell_lines.append('')



from abc import ABC

import numpy as np
from .numeric import NumericIndicator

try:
    import torch
except ImportError:
    torch = None


class IndexedIndicator(NumericIndicator, ABC):
    def __init__(self, name: str):
        super().__init__(name=name, is_print=False)
        self._values = []
        self._indexes = []

    def clear(self):
        self._values = []
        self._indexes = []

    def collect_value(self, value):
        if type(value) == tuple:
            assert len(value) == 2
            if type(value[0]) == int:
                self._indexes.append(value[0])
                self._values.append(value[1])
            else:
                assert type(value[0]) == list
                assert len(value[0]) == len(value[1])
                self._indexes += value[0]
                self._values += value[1]
        else:
            assert type(value) == list
            self._indexes += [v[0] for v in value]
            self._values += [v[1] for v in value]

    def is_empty(self) -> bool:
        return len(self._values) == 0

    def get_mean(self) -> float:
        return float(np.mean(self._values))

    def get_index_mean(self):
        summary = {}
        for ind, values in zip(self._indexes, self._values):
            if ind not in summary:
                summary[ind] = []
            summary[ind].append(values)

        indexes = []
        means = []
        for ind, values in summary.items():
            indexes.append(ind)
            means.append(float(np.mean(values)))

        return indexes, means


class IndexedScalar(IndexedIndicator):
    def get_histogram(self):
        return None

    def copy(self, key: str):
        return IndexedScalar(key)



import torch

from labml.configs import BaseConfigs, hyperparams, option


class DeviceInfo:
    def __init__(self, *,
                 use_cuda: bool,
                 cuda_device: int):
        self.use_cuda = use_cuda
        self.cuda_device = cuda_device
        self.cuda_count = torch.cuda.device_count()

        self.is_cuda = self.use_cuda and torch.cuda.is_available()
        if not self.is_cuda:
            self.device = torch.device('cpu')
        else:
            if self.cuda_device < self.cuda_count:
                self.device = torch.device('cuda', self.cuda_device)
            else:
                self.device = torch.device('cuda', self.cuda_count - 1)

    def __str__(self):
        if not self.is_cuda:
            return "CPU"

        if self.cuda_device < self.cuda_count:
            return f"GPU:{self.cuda_device} - {torch.cuda.get_device_name(self.cuda_device)}"
        else:
            return (f"GPU:{self.cuda_count - 1}({self.cuda_device}) "
                    f"- {torch.cuda.get_device_name(self.cuda_count - 1)}")


class DeviceConfigs(BaseConfigs):
    cuda_device: int = 0
    use_cuda: bool = True

    device_info: DeviceInfo

    device: torch.device

    def __init__(self):
        super().__init__(_primary='device')


@option(DeviceConfigs.device)
def _device(c: DeviceConfigs):
    return c.device_info.device


hyperparams(DeviceConfigs.cuda_device, DeviceConfigs.use_cuda,
            is_hyperparam=False)


@option(DeviceConfigs.device_info)
def _device_info(c: DeviceConfigs):
    return DeviceInfo(use_cuda=c.use_cuda,
                      cuda_device=c.cuda_device)



import numpy as np
import torch
from labml import monit, lab
from labml import tracker
from torch.utils.data import Dataset

from labml_samples.pytorch.stocks import CandleIdx
from labml_samples.pytorch.stocks.build_numpy_cache import build_cache


def normalize_by_previous_day(packs: torch.Tensor):
    mean_price = packs[:, :, 0:CandleIdx.open_close].mean(dim=-1).mean(dim=-1)
    mean_volume = packs[:, :, CandleIdx.volume].mean(dim=-1)
    mean_price = mean_price.clamp_min(1e-6)
    mean_volume = mean_volume.clamp_min(1e-6)

    data = packs[1:, :, :] / mean_price[:-1].view(-1, 1, 1)
    data[:, :, CandleIdx.volume] = packs[1:, :, CandleIdx.volume] / mean_volume[:-1].view(-1, 1)

    return data.transpose_(0, 1)


def calculate_moving_average(data: torch.Tensor, start: int, end: int):
    cat = []
    pad = data.new_zeros(1, *data.shape[1:])
    cat.append(pad)
    if start <= 0:
        # pad = data.new_zeros(-start, *data.shape[1:])
        pad = data[:1].repeat(-start, *[1 for _ in data.shape[1:]])
        cat.append(pad)

    if start > 0:
        data = data[start:]
    if end < 0:
        data = data[:end]
    cat.append(data)

    if end > 0:
        # pad = data.new_zeros(end, *data.shape[1:])
        pad = data[-1:].repeat(end, *[1 for _ in data.shape[1:]])
        cat.append(pad)

    data = torch.cat(cat, dim=0)
    cum = data.cumsum(dim=0)

    range_ = end - start + 1

    avg = cum[range_:] - cum[:-range_]
    avg /= range_

    return avg


class MinutelyDataset(Dataset):
    def __init__(self, dates: np.ndarray, packets: torch.Tensor):
        self.data = normalize_by_previous_day(packets)
        self.data.log_()
        mean = self.data[:, :, 0:CandleIdx.open_close].mean(dim=-1)
        self.reference = calculate_moving_average(mean, -4, 0)
        self.target = calculate_moving_average(mean, 1, 5)
        self.strike_low = self.data[:, :, CandleIdx.low]
        self.strike_high = self.data[:, :, CandleIdx.high]
        self.strike_low[:-1] = self.data[1:, :, CandleIdx.low]
        self.strike_high[:-1] = self.data[1:, :, CandleIdx.high]

        self.data.transpose_(0, 1)
        self.reference.transpose_(0, 1).unsqueeze_(-1)
        self.target.transpose_(0, 1).unsqueeze_(-1)
        self.strike_low.transpose_(0, 1).unsqueeze_(-1)
        self.strike_high.transpose_(0, 1).unsqueeze_(-1)
        self.y = self.target - self.reference

        self.dates = dates[1:]

        self.price_std = self.data[:, :, 0:CandleIdx.prices].std().item()
        self.price_mean = self.data[:, :, 0:CandleIdx.prices].mean().item()
        self.volume_std = self.data[:, :, CandleIdx.volume].std().item()
        self.volume_mean = self.data[:, :, CandleIdx.volume].mean().item()
        self.y_std = self.y.std().item()
        self.y_mean = self.y.mean().item()

        tracker.set_tensor("ref.*", is_once=True)
        tracker.set_tensor("target.*", is_once=True)
        tracker.set_tensor("strike_low.*", is_once=True)
        tracker.set_tensor("strike_high.*", is_once=True)

    def __len__(self):
        return self.data.shape[0]

    def __getitem__(self, idx):
        if torch.is_tensor(idx):
            idx = idx.tolist()

        return {'data': self.data[idx],
                'target': self.y[idx]}

    def save_artifacts(self):
        tracker.save({'target.': self.y,
                      'ref.': self.reference,
                      'strike_low.': self.strike_low,
                      'strike_high.': self.strike_high})


class MinutelyData:
    def __init__(self, validation_dates: int, skip_cache: bool = False):
        self.validation_dates = validation_dates

        dates_cache_path = lab.get_data_path() / 'dates.npy'
        packets_cache_path = lab.get_data_path() / 'packets.npy'

        if skip_cache or not dates_cache_path.exists() or not packets_cache_path.exists():
            with monit.section('Build cache'):
                build_cache()

        with monit.section("Cache"):
            self.dates = np.load(str(dates_cache_path))
            self.packets = torch.tensor(np.load(str(packets_cache_path)), dtype=torch.float)

    def train_dataset(self):
        return MinutelyDataset(self.dates[:-self.validation_dates], self.packets[:-self.validation_dates])

    def valid_dataset(self):
        return MinutelyDataset(self.dates[-self.validation_dates:], self.packets[-self.validation_dates:])



"""
---
title: Configurable optimizer module
summary: This implements a configurable module for optimizers.
---
"""

from typing import Tuple

import torch

from labml.configs import BaseConfigs, option, meta_config
from labml_nn.optimizers import WeightDecay


class OptimizerConfigs(BaseConfigs):
    optimizer: torch.optim.Adam

    weight_decay_obj: WeightDecay
    weight_decouple: bool = True
    weight_decay: float = 0.0
    weight_decay_absolute: bool = False
    optimized_adam_update: bool = True

    parameters: any

    learning_rate: float = 0.01
    betas: Tuple[float, float] = (0.9, 0.999)
    eps: float = 1e-08

    momentum: float = 0.5
    amsgrad: bool = False
    warmup: int = 0
    degenerated_to_sgd: bool = True
    rectify: bool = True
    d_model: int
    step_factor: int = 1024

    def __init__(self):
        super().__init__(_primary='optimizer')


meta_config(OptimizerConfigs.parameters)


@option(OptimizerConfigs.weight_decay_obj, 'L2')
def _weight_decay(c: OptimizerConfigs):
    return WeightDecay(c.weight_decay, c.weight_decouple, c.weight_decay_absolute)


@option(OptimizerConfigs.optimizer, 'SGD')
def _sgd_optimizer(c: OptimizerConfigs):
    return torch.optim.SGD(c.parameters, c.learning_rate, c.momentum)


@option(OptimizerConfigs.optimizer, 'Adam')
def _adam_optimizer(c: OptimizerConfigs):
    if c.amsgrad:
        from labml_nn.optimizers.amsgrad import AMSGrad
        return AMSGrad(c.parameters,
                       lr=c.learning_rate, betas=c.betas, eps=c.eps,
                       optimized_update=c.optimized_adam_update,
                       weight_decay=c.weight_decay_obj, amsgrad=c.amsgrad)
    else:
        from labml_nn.optimizers.adam import Adam
        return Adam(c.parameters,
                    lr=c.learning_rate, betas=c.betas, eps=c.eps,
                    optimized_update=c.optimized_adam_update,
                    weight_decay=c.weight_decay_obj)


@option(OptimizerConfigs.optimizer, 'AdamW')
def _adam_warmup_optimizer(c: OptimizerConfigs):
    from labml_nn.optimizers.adam_warmup import AdamWarmup
    return AdamWarmup(c.parameters,
                      lr=c.learning_rate, betas=c.betas, eps=c.eps,
                      weight_decay=c.weight_decay_obj, amsgrad=c.amsgrad, warmup=c.warmup)


@option(OptimizerConfigs.optimizer, 'RAdam')
def _radam_optimizer(c: OptimizerConfigs):
    from labml_nn.optimizers.radam import RAdam
    return RAdam(c.parameters,
                 lr=c.learning_rate, betas=c.betas, eps=c.eps,
                 weight_decay=c.weight_decay_obj, amsgrad=c.amsgrad,
                 degenerated_to_sgd=c.degenerated_to_sgd)


@option(OptimizerConfigs.optimizer, 'AdaBelief')
def _ada_belief_optimizer(c: OptimizerConfigs):
    from labml_nn.optimizers.ada_belief import AdaBelief
    return AdaBelief(c.parameters,
                     lr=c.learning_rate, betas=c.betas, eps=c.eps,
                     weight_decay=c.weight_decay_obj, amsgrad=c.amsgrad,
                     degenerated_to_sgd=c.degenerated_to_sgd,
                     rectify=c.rectify)


@option(OptimizerConfigs.optimizer, 'Noam')
def _noam_optimizer(c: OptimizerConfigs):
    from labml_nn.optimizers.noam import Noam
    return Noam(c.parameters,
                lr=c.learning_rate, betas=c.betas, eps=c.eps,
                weight_decay=c.weight_decay_obj, amsgrad=c.amsgrad, warmup=c.warmup,
                d_model=c.d_model)



import signal

from labml import logger
from labml.logger import Text


class DelayedKeyboardInterrupt:
    """
    ### Capture `KeyboardInterrupt` and fire it later
    """

    def __init__(self):
        self.signal_received = None

    def __enter__(self):
        self.signal_received = None
        # Start capturing
        self.old_handler = signal.signal(signal.SIGINT, self.handler)

    def handler(self, sig, frame):
        # Pass second interrupt without delaying
        if self.signal_received is not None:
            self.old_handler(*self.signal_received)
            return

        # Store the interrupt signal for later
        self.signal_received = (sig, frame)
        logger.log([('\nSIGINT received. Delaying KeyboardInterrupt.',
                     Text.danger)])

    def __exit__(self, exc_type, exc_val, exc_tb):
        # Reset handler
        signal.signal(signal.SIGINT, self.old_handler)

        # Pass on any captured interrupt signals
        if self.signal_received is not None:
            self.old_handler(*self.signal_received)



from typing import Callable


class EvalFunction:
    func: Callable
    config_name: str

    def __init__(self, func, *,
                 config_name: str):
        self.func = func
        self.config_name = config_name



from pathlib import PurePath
from typing import List, Tuple

import numpy as np

from labml.internal.analytics.indicators import IndicatorClass, Indicator, Run, IndicatorCollection, \
    StepSelect
from labml.internal.analytics.sqlite import SQLiteAnalytics
from labml.internal.analytics.tensorboard import TensorBoardAnalytics

_RUNS = {}

_TENSORBOARD = {}

_SQLITE = {}

_NUMPY_ARRAYS = {}


def get_run(uuid: str) -> Run:
    if uuid not in _RUNS:
        _RUNS[uuid] = Run(uuid)

    return _RUNS[uuid]


def get_name(ind: Indicator) -> List[str]:
    run = get_run(ind.uuid)
    return [run.name, ind.uuid[-5:], run.run_info.comment, ind.key]


def get_tensorboard_data(indicator: Indicator):
    run = get_run(indicator.uuid)

    if indicator.uuid not in _TENSORBOARD:
        _TENSORBOARD[indicator.uuid] = TensorBoardAnalytics(run.run_info.tensorboard_log_path)

    tb: TensorBoardAnalytics = _TENSORBOARD[indicator.uuid]
    try:
        tb.load()
    except FileNotFoundError:
        return None

    key = indicator.key
    if indicator.is_distribution and indicator.is_mean:
        key = f'{key}.mean'

    try:
        tensor = tb.tensor(key, indicator.select.start, indicator.select.end)
    except KeyError:
        return None

    if indicator.is_distribution and not indicator.is_mean:
        data = tb.summarize_compressed_histogram(tensor)
    else:
        data = tb.summarize_scalars(tensor)

    return data


def _get_sqlite_scalar_data(sqlite: SQLiteAnalytics, key: str, select: StepSelect):
    data = sqlite.scalar(key, select.start, select.end)
    if not data:
        return None

    data = sqlite.summarize_scalars(data)

    return data


def get_sqlite_data(indicator: Indicator):
    run = get_run(indicator.uuid)

    if indicator.uuid not in _SQLITE:
        _SQLITE[indicator.uuid] = SQLiteAnalytics(run.run_info.sqlite_path)

    sqlite: SQLiteAnalytics = _SQLITE[indicator.uuid]

    if indicator.is_distribution:
        return _get_sqlite_scalar_data(sqlite, f"{indicator.key}.mean", indicator.select)
    elif indicator.is_scalar:
        return _get_sqlite_scalar_data(sqlite, indicator.key, indicator.select)
    else:
        return None


_PREFERRED_DB = 'tensorboard'


def set_preferred_db(db: str):
    global _PREFERRED_DB
    _PREFERRED_DB = db


def get_indicator_data(indicator: Indicator):
    if _PREFERRED_DB == 'tensorboard':
        data = get_tensorboard_data(indicator)
        if data is None:
            data = get_sqlite_data(indicator)
    else:
        data = get_sqlite_data(indicator)
        if data is None:
            data = get_tensorboard_data(indicator)

    return data


def get_indicators_data(indicators: IndicatorCollection):
    series = []
    names = []
    for i, ind in enumerate(indicators):
        d = get_indicator_data(ind)
        if d is not None:
            series.append(d)
            names.append(get_name(ind))

    return series, names


def _get_numpy_array(path: PurePath):
    path = str(path)
    if path not in _NUMPY_ARRAYS:
        _NUMPY_ARRAYS[path] = np.load(path)

    return _NUMPY_ARRAYS[path]


def get_artifact_files(indicator: Indicator):
    run = get_run(indicator.uuid)

    if indicator.uuid not in _SQLITE:
        _SQLITE[indicator.uuid] = SQLiteAnalytics(run.run_info.sqlite_path)

    sqlite: SQLiteAnalytics = _SQLITE[indicator.uuid]

    if indicator.class_ != IndicatorClass.tensor:
        return None

    data = sqlite.tensor(indicator.key, indicator.select.start, indicator.select.end)

    if not data:
        return None

    return data


def _get_condensed_steps(series: List[List[Tuple[int, any]]], limit: int):
    steps = {}
    step_lookups = []

    series = [s for s in series if len(s) != 1 or s[0][0] != -1]

    for s in series:
        lookup = {}
        for i, c in enumerate(s):
            steps[c[0]] = steps.get(c[0], 0) + 1
            lookup[c[0]] = i
        step_lookups.append(lookup)

    steps = [k for k, v in steps.items() if v == len(series)]
    if len(steps) == 0:
        return [], step_lookups

    steps = sorted(steps)
    last = steps[-1]
    interval = max(1, last // (limit - 1))

    condensed = [steps[0]]
    for s in steps[1:]:
        if s - condensed[-1] >= interval:
            condensed.append(s)
    if condensed[-1] != steps[-1]:
        condensed.append(steps[-1])
    return condensed, step_lookups


def get_artifacts_data(indicators: IndicatorCollection, limit: int = 100):
    series = []
    names = []
    series_inds = []
    for i, ind in enumerate(indicators):
        d = get_artifact_files(ind)
        if d is not None:
            series.append(d)
            series_inds.append(ind)
            names.append(get_name(ind))

    steps, step_lookups = _get_condensed_steps(series, limit)

    data_series = []
    for si, s in enumerate(series):
        ind: Indicator = series_inds[si]
        run: Run = get_run(ind.uuid)
        if ind.props.get('is_once', False):
            assert s[0][0] == -1
            filename = s[0][1]
            data = _get_numpy_array(run.run_info.artifacts_folder / filename)
        else:
            data = []
            for step in steps:
                filename = s[step_lookups[si][step]][1]
                data.append((step, _get_numpy_array(run.run_info.artifacts_folder / filename)))
        data_series.append(data)

    return data_series, names



import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import torch.utils.data
from torchvision import datasets, transforms

from labml import experiment, lab, tracker, logger
from labml.configs import BaseConfigs, option


class Net(nn.Module):
    def __init__(self):
        super().__init__()
        self.conv1 = nn.Conv2d(1, 20, 5, 1)
        self.conv2 = nn.Conv2d(20, 50, 5, 1)
        self.fc1 = nn.Linear(4 * 4 * 50, 500)
        self.fc2 = nn.Linear(500, 10)

    def forward(self, x):
        x = F.relu(self.conv1(x))
        x = F.max_pool2d(x, 2, 2)
        x = F.relu(self.conv2(x))
        x = F.max_pool2d(x, 2, 2)
        x = x.view(-1, 4 * 4 * 50)
        x = F.relu(self.fc1(x))
        return self.fc2(x)


class LoaderConfigs(BaseConfigs):
    train_batch_size: int = 64
    valid_batch_size: int = 1000

    train_loader: torch.utils.data.DataLoader
    valid_loader: torch.utils.data.DataLoader


def _data_loader(is_train, batch_size):
    return torch.utils.data.DataLoader(
        datasets.MNIST(str(lab.get_data_path()),
                       train=is_train,
                       download=True,
                       transform=transforms.Compose([
                           transforms.ToTensor(),
                           transforms.Normalize((0.1307,), (0.3081,))
                       ])),
        batch_size=batch_size, shuffle=True)


@option([LoaderConfigs.train_loader, LoaderConfigs.valid_loader])
def data_loaders(c: LoaderConfigs):
    train_data = _data_loader(True, c.train_batch_size)
    valid_data = _data_loader(False, c.valid_batch_size)

    return train_data, valid_data


class Configs(LoaderConfigs):
    epochs: int = 10

    use_cuda: bool = True
    seed: int = 5

    device: any

    model: nn.Module

    learning_rate: float = 0.01
    momentum: float = 0.5
    optimizer: optim.Adam = 'adam_optimizer'

    train_log_interval = 10

    def train(self):
        self.model.train()
        for batch_idx, (data, target) in enumerate(self.train_loader):
            data, target = data.to(self.device), target.to(self.device)

            self.optimizer.zero_grad()
            output = self.model(data)
            loss = F.cross_entropy(output, target)
            loss.backward()
            self.optimizer.step()

            # **✨ Increment the global step**
            tracker.add_global_step()
            # **✨ Store stats in the tracker**
            tracker.save({'loss.train': loss})

            #
            if batch_idx % self.train_log_interval == 0:
                # **✨ Save added stats**
                tracker.save()

    def validate(self):
        self.model.eval()
        valid_loss = 0
        correct = 0
        with torch.no_grad():
            for data, target in self.valid_loader:
                data, target = data.to(self.device), target.to(self.device)

                output = self.model(data)
                valid_loss += F.cross_entropy(output, target, reduction='sum').item()
                pred = output.argmax(dim=1, keepdim=True)
                correct += pred.eq(target.view_as(pred)).sum().item()

        valid_loss /= len(self.valid_loader.dataset)
        valid_accuracy = 100. * correct / len(self.valid_loader.dataset)

        # **Save stats**
        tracker.save({'loss.valid': valid_loss, 'accuracy.valid': valid_accuracy})

    def run(self):
        for epoch in range(1, self.epochs + 1):
            self.train()
            self.validate()
            logger.log()


@option(Configs.device)
def get_device(c: Configs):
    is_cuda = c.use_cuda and torch.cuda.is_available()
    if not is_cuda:
        return torch.device("cpu")
    else:
        return torch.device(f"cuda:0")


@option(Configs.model)
def model(c: Configs):
    return Net().to(c.device)


@option(Configs.optimizer)
def sgd_optimizer(c: Configs):
    return optim.SGD(c.model.parameters(), lr=c.learning_rate, momentum=c.momentum)


@option(Configs.optimizer)
def adam_optimizer(c: Configs):
    return optim.Adam(c.model.parameters(), lr=c.learning_rate)


def main():
    conf = Configs()
    experiment.create(name='configs')
    experiment.configs(conf, {'optimizer': 'sgd_optimizer'})

    torch.manual_seed(conf.seed)

    with experiment.start():
        conf.run()

    # save the model
    experiment.save_checkpoint()


if __name__ == '__main__':
    main()



import numpy as np
import torch

from labml.configs import BaseConfigs, option


class SetSeed:
    def __init__(self, seed: int):
        self.seed = seed

    def __call__(self):
        torch.manual_seed(self.seed)
        np.random.seed(self.seed)


class SeedConfigs(BaseConfigs):
    seed: int = 5

    set = '_set_seed'


@option(SeedConfigs.set)
def _set_seed(c: SeedConfigs):
    return SetSeed(c.seed)



from typing import List, Union, Tuple, Callable, Optional

from IPython.core.display import display, HTML

from labml.internal.logger.destinations import Destination
from labml.internal.util import is_kaggle
from labml.internal.util.colors import StyleCode

get_ipython: Callable


class IpynbDestination(Destination):
    def __init__(self):
        self.__last_handle = None
        self.__last_id = 1
        self.__cell_lines = []
        self.__cell_count = 0

    def is_same_cell(self):
        try:
            cells = get_ipython().ev('len(In)')
        except NameError:
            return False

        if cells == self.__cell_count:
            return True

        self.__cell_count = cells
        self.__cell_lines = []
        self.__last_handle = None

        return False

    @staticmethod
    def html_code(text: str, color: List[StyleCode] or StyleCode or None):
        """
        ### Add ansi color codes
        """
        if text == '\n':
            assert color is None
            return text

        if color is None:
            return text
        elif isinstance(color, list):
            open_tags = ''.join([c.html_open(text) for c in color])
            close_tags = ''.join([c.html_close(text) for c in reversed(color)])
        else:
            open_tags = color.html_open(text)
            close_tags = color.html_close(text)

        return open_tags + text + close_tags

    def log(self, parts: List[Union[str, Tuple[str, Optional[StyleCode]]]], *,
            is_new_line: bool,
            is_reset: bool):
        tuple_parts = []
        for p in parts:
            if type(p) == str:
                text = p
                style = None
            else:
                text, style = p
            lines = text.split('\n')
            for line in lines[:-1]:
                tuple_parts.append((line, style))
                tuple_parts.append(('\n', None))
            tuple_parts.append((lines[-1], style))

        coded = [self.html_code(text, color) for text, color in tuple_parts]

        text = "".join(coded)
        lines = text.split('\n')

        if is_kaggle():
            attrs = 'style="color: #444; overflow-x: scroll;"'
        else:
            attrs = 'style="overflow-x: scroll;"'

        if self.is_same_cell():
            if coded:
                last = self.__cell_lines.pop()
                if is_reset:
                    self.__cell_lines += lines
                else:
                    self.__cell_lines += [last + lines[0]] + lines[1:]
            text = '\n'.join(self.__cell_lines)
            html = HTML(f"<pre {attrs}>{text}</pre>")
            self.__last_handle.update(html)
        else:
            self.__cell_lines = lines
            text = '\n'.join(self.__cell_lines)
            html = HTML(f"<pre {attrs}>{text}</pre>")
            self.__last_handle = display(html, display_id=self.__last_id)
            self.__last_id += 1

        # print(len(self.__cell_lines), self.__cell_lines[-1], is_new_line)
        if is_new_line:
            self.__cell_lines.append('')



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from . import Experiment


class ExperimentWatcher:
    def __init__(self, exp: 'Experiment'):
        self.exp = exp

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None:
            self.exp.finish('completed')
        elif exc_type == KeyboardInterrupt:
            self.exp.finish('interrupted')
        else:
            self.exp.finish('crashed', str(exc_val))



import math
import time
from typing import TYPE_CHECKING, List

from labml.logger import Text
from ..tracker import tracker_singleton as tracker

if TYPE_CHECKING:
    from ..monitor import Monitor


class Section:
    r"""
        Note:
            You should use :meth:`labml.logger.section` to create sections
    """

    def __init__(self, *,
                 monitor: 'Monitor',
                 name: str,
                 is_silent: bool,
                 is_timed: bool,
                 is_partial: bool,
                 is_children_silent: bool,
                 total_steps: float):
        self.is_children_silent = is_children_silent
        self._monitor = monitor
        self._name = name
        self.is_silent = is_silent
        self._is_timed = is_timed
        self._is_partial = is_partial
        self._total_steps = total_steps

        self._state = 'none'
        self._has_entered_ever = False

        self._start_time = 0
        self._end_time = -1
        self._progress = 0.
        self._start_progress = 0
        self._end_progress = 0
        self._is_parented = False

        self.is_successful = True
        self.message = None

    @property
    def name(self):
        return self._name

    def get_estimated_time(self) -> float:
        raise NotImplementedError()

    def __enter__(self):
        self._state = 'entered'
        self._has_entered_ever = True
        self.is_successful = True

        if not self._is_partial:
            self._progress = 0

        self._start_progress = self._progress

        if self._is_timed:
            self._start_time = time.time()

        self._monitor.section_enter(self)

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_val is not None:
            self.is_successful = False
        self._state = 'exited'
        if self._is_timed:
            self._end_time = time.time()

        if not self._is_partial:
            self._progress = 1.

        self._end_progress = self._progress

        self.track_progress()

        self._monitor.section_exit(self)

    def track_progress(self):
        pass

    def log(self):
        raise NotImplementedError()

    def progress(self, steps):
        old_progress = self._progress
        self._progress = steps / self._total_steps

        if self.is_silent:
            return False

        if math.floor(self._progress * 100) != math.floor(old_progress * 100):
            return True
        else:
            return False

    @property
    def is_parented(self):
        return self._is_parented

    def make_parent(self):
        self._is_parented = True


class OuterSection(Section):
    def __init__(self, *,
                 monitor: 'Monitor',
                 name: str,
                 is_silent: bool,
                 is_timed: bool,
                 is_partial: bool,
                 is_children_silent: bool,
                 is_new_line: bool,
                 total_steps: float,
                 level: int):
        if is_partial:
            raise RuntimeError("Only sections within the loop can be partial.")

        super().__init__(monitor=monitor,
                         name=name,
                         is_silent=is_silent,
                         is_timed=is_timed,
                         is_partial=is_partial,
                         is_children_silent=is_children_silent,
                         total_steps=total_steps)

        self._level = level
        self._is_new_line = is_new_line

    def get_estimated_time(self):
        if self._state == 'entered':
            if self._progress == 0.:
                return time.time() - self._start_time
            else:
                return (time.time() - self._start_time) / self._progress
        else:
            return self._end_time - self._start_time

    def log(self):
        if self.is_silent:
            return

        if self._state == 'none':
            return

        parts = [("  " * self._level + f"{self._name}", None)]

        if self._state == 'entered':
            if self._progress == 0.:
                parts.append("...")
            else:
                parts.append((f" {math.floor(self._progress * 100) :4.0f}%", Text.meta2))
        else:
            if self.is_successful:
                parts.append(("...[DONE]", Text.success))
            else:
                parts.append(("...[FAIL]", Text.danger))

        if self._is_timed and self._progress > 0.:
            duration_ms = 1000 * self.get_estimated_time()
            parts.append((f"\t{duration_ms :,.2f}ms",
                          Text.meta))

        if self.message is not None:
            parts.append((f"\t{self.message}", Text.value))

        if self._state != 'entered' and self._is_new_line:
            parts.append(("\n", None))

        return parts


class LoopingSection(Section):
    def __init__(self, *,
                 monitor: 'Monitor',
                 name: str,
                 is_silent: bool,
                 is_timed: bool,
                 is_track: bool,
                 is_partial: bool,
                 total_steps: float,
                 parents: List[str]):
        super().__init__(monitor=monitor,
                         name=name,
                         is_silent=is_silent,
                         is_timed=is_timed,
                         is_partial=is_partial,
                         is_children_silent=False,
                         total_steps=total_steps)
        self._beta_pow = 1.
        self._beta = 0.9
        self._estimated_time = 0.
        self._time_length = 7
        self._last_end_time = -1.
        self._last_start_time = -1.
        self._last_step_time = 0.
        self._last_est_time = 0
        self._is_track = is_track
        self._parents = parents

    @property
    def is_child(self) -> bool:
        return len(self._parents) > 0

    def track_progress(self):
        name = '.'.join(self._parents + [self._name])
        tracker().store(f'time.{name}', self._calc_estimated_time())

    def get_estimated_time(self):
        et = self._estimated_time * self._beta
        et += (1 - self._beta) * self._last_step_time
        return et / (1 - self._beta_pow * self._beta)

    def _calc_estimated_time(self):
        if self._state != 'entered':
            if self._last_end_time == self._end_time:
                return self.get_estimated_time()
            end_time = self._end_time
            end_progress = self._end_progress
            self._last_end_time = self._end_time
        else:
            end_time = time.time()
            end_progress = self._progress

        if end_progress - self._start_progress < 1e-6:
            return self.get_estimated_time()

        current_estimate = ((end_time - self._start_time) /
                            (end_progress - self._start_progress))

        if self._last_start_time == self._start_time and end_time < self._last_est_time + 2:
            # print(current_estimate)
            self._last_step_time = current_estimate
        else:
            if self._last_step_time >= 0.:
                self._beta_pow *= self._beta
                self._estimated_time *= self._beta
                self._estimated_time += (1 - self._beta) * self._last_step_time
            # print(self._last_step_time, current_estimate)
            self._last_step_time = current_estimate
            self._last_start_time = self._start_time
            self._last_est_time = end_time

        return self.get_estimated_time()

    def log(self):
        if self.is_silent:
            return []

        if self._state == 'none':
            return []

        parts = [(f"{self._name}:", None)]
        color = None

        if not self.is_successful:
            color = Text.danger

        if self._progress == 0.:
            parts.append(("  ...", Text.subtle))
        else:
            parts.append((f"{math.floor(self._progress * 100) :4.0f}%",
                          color or Text.subtle))

        if self._is_timed:
            duration_ms = 1000 * self._calc_estimated_time()
            s = f" {duration_ms:,.0f}ms  "
            tl = len(s)
            if tl > self._time_length:
                self._time_length = tl
            else:
                s = (" " * (self._time_length - tl)) + s

            parts.append((s, color or Text.meta))

        return parts



from collections import deque
from pathlib import PurePath
from typing import Dict, Optional

import numpy as np
from torch.utils.tensorboard import SummaryWriter

from . import TensorboardWriter


class PyTorchTensorboardWriter(TensorboardWriter):
    __writer: Optional[SummaryWriter]

    def __init__(self, log_path: PurePath):
        super().__init__()

        self.__log_path = log_path
        self.__writer = None

    def __connect(self):
        if self.__writer is not None:
            return

        self.__writer = SummaryWriter(str(self.__log_path))

    def hparams(self, hparams: Dict[str, any]):
        self.__connect()

    def add_scalar(self, key: str, value: any, global_step: int):
        self.__connect()
        self.__writer.add_scalar(key, value, global_step)

    def add_histogram(self, key: str, value: any, global_step: int):
        self.__connect()
        if isinstance(value, list) or isinstance(value, deque):
            value = np.array(value)
        self.__writer.add_histogram(key, value, global_step)

    def add_image(self, key: str, img: any, global_step: int):
        self.__connect()
        if len(img.shape) == 4:
            for i in range(len(img)):
                self.__writer.add_image(key, img[i], global_step)
        else:
            self.__writer.add_image(key, img, global_step)



from typing import Dict, Optional

import tensorflow as tf

from labml import tracker, logger

_MAP = {
    'size': None,
    'batch': None,
    'loss': 'loss.train',
    'accuracy': 'accuracy.train',
    'val_loss': 'loss.valid',
    'val_accuracy': 'accuracy.valid'
}


class LabMLKerasCallback(tf.keras.callbacks.Callback):
    def __init__(self, save_batch_frequency: int = 1):
        super().__init__()
        self.save_batch_frequency = save_batch_frequency

    @staticmethod
    def _parse_logs(logs: Optional[Dict[str, any]]):
        data = {}
        if logs is None:
            return data
        for k, v in logs.items():
            if k in _MAP:
                k = _MAP[k]
            if k is None:
                continue

            data[k] = v

        return data

    def on_epoch_end(self, epoch, logs=None):
        tracker.save(self._parse_logs(logs))
        logger.log()

    def on_train_batch_end(self, batch, logs=None):
        tracker.add_global_step()
        tracker.add(self._parse_logs(logs))
        if batch % self.save_batch_frequency == 0:
            tracker.save()



from matplotlib.axes import Axes


class MatPlotLibAnalytics:
    def render_density(self, ax: Axes, data, color, name, *,
                       levels=5,
                       line_width=1,
                       alpha=0.6):
        ln = ax.plot(data[:, 0], data[:, 5],
                     lw=line_width,
                     color=color,
                     alpha=1,
                     label=name)

        for i in range(1, levels):
            ax.fill_between(
                data[:, 0],
                data[:, 5 - i],
                data[:, 5 + i],
                color=color,
                lw=0,
                alpha=alpha ** i)

        return ln



import math
import time
from typing import Optional, Dict, TYPE_CHECKING, List, Collection

from labml.internal.monitor.sections import LoopingSection
from labml.logger import Text
from ..tracker import tracker_singleton as tracker

if TYPE_CHECKING:
    from ..monitor import Monitor


class Loop:
    def __init__(self, iterator: Collection, *,
                 monitor: 'Monitor',
                 is_track: bool,
                 is_print_iteration_time: bool):
        """
        Creates an iterator with a range `iterator`.

        See example for usage.
        """
        self.iterator = iterator
        self._start_time = 0.
        self._iter_start_time = 0.
        self._iter_start_step = 0
        self._init_time = 0.
        self._iter_time = 0.
        self._beta_pow = 1.
        self._beta = 0.9
        self.monitor = monitor
        self.__global_step: Optional[int] = None
        self.__looping_sections: Dict[str, LoopingSection] = {}
        self._is_print_iteration_time = is_print_iteration_time
        self._is_track = is_track
        self._started = False

    def __iter__(self):
        self.iterator_iter = iter(self.iterator)
        self._start_time = time.time()
        self._init_time = 0.
        self._iter_time = 0.
        self._iter_start_step = tracker().global_step
        self._started = False
        self.monitor.start_loop()
        return self

    def __next__(self):
        try:
            next_value = next(self.iterator_iter)
        except StopIteration as e:
            self.monitor.finish_loop()
            raise e

        now = time.time()
        if not self._started:
            self.__init_time = now - self._start_time
        else:
            self._beta_pow *= self._beta
            self._iter_time *= self._beta
            self._iter_time += (1 - self._beta) * (now - self._iter_start_time)
            if self._is_track:
                tracker().store('time.loop',
                                (now - self._iter_start_time) /
                                max(1, tracker().global_step - self._iter_start_step))

        self._iter_start_time = now
        self._iter_start_step = tracker().global_step
        self._started = True

        tracker().loop_count(next_value)

        return next_value

    def log_progress(self):
        """
        Show progress
        """
        now = time.time()
        spent = now - self._start_time

        if not math.isclose(self._iter_time, 0.):
            estimate = self._iter_time / (1 - self._beta_pow)
        else:
            estimate = sum([s.get_estimated_time() for s in self.__looping_sections.values() if not s.is_child])

        total_time = estimate * len(self.iterator) + self._init_time
        total_time = max(total_time, spent)
        remain = total_time - spent

        remain /= 60
        spent /= 60
        estimate *= 1000

        spent_h = int(spent // 60)
        spent_m = int(spent % 60)
        remain_h = int(remain // 60)
        remain_m = int(remain % 60)

        to_print = [("  ", None)]
        if self._is_print_iteration_time:
            to_print.append((f"{estimate:,.0f}ms", Text.meta))
        to_print.append((f"{spent_h:3d}:{spent_m:02d}m/{remain_h:3d}:{remain_m:02d}m  ", Text.meta2))

        return to_print

    def get_section(self, *, name: str,
                    is_silent: bool,
                    is_timed: bool,
                    is_partial: bool,
                    total_steps: float,
                    parents: List[str]):
        key = '.'.join(parents + [name])
        if key not in self.__looping_sections:
            self.__looping_sections[key] = LoopingSection(monitor=self.monitor,
                                                          name=name,
                                                          is_silent=is_silent,
                                                          is_timed=is_timed,
                                                          is_track=self._is_track,
                                                          is_partial=is_partial,
                                                          total_steps=total_steps,
                                                          parents=parents)
        return self.__looping_sections[key]

    def log_sections(self):
        parts = []
        for name, section in self.__looping_sections.items():
            parts += section.log()

        return parts



from typing import Optional

import altair as alt
import numpy as np

try:
    import torch
except ImportError:
    torch = None


def data_to_table(data, low: Optional[float], high: Optional[float]):
    table = []

    if torch is not None and isinstance(data, torch.Tensor):
        data = data.detach().cpu().numpy()

    data: np.ndarray = data.ravel()

    for i in range(data.shape[0]):
        v = data[i]
        if low is not None and v < low:
            continue
        if high is not None and v > high:
            continue
        table.append({'value': v})

    return alt.Data(values=table)


def render(table: alt.Data, *,
           height: int,
           width: int,
           height_minimap: int):
    brush = alt.selection_interval(encodings=['x'])

    base = alt.Chart(table).mark_bar().encode(
        y='count():Q'
    )
    detail = base.encode(
        alt.X('value:Q',
              bin=alt.Bin(maxbins=30, extent=brush),
              scale=alt.Scale(domain=brush)
              )
    ).properties(
        width=width,
        height=height
    )
    minimap = base.encode(
        alt.X('value:Q', bin=alt.Bin(maxbins=30)),
    ).add_selection(brush).properties(width=width, height=height_minimap)

    return detail & minimap



import time
from typing import Dict

import numpy as np
from labml.internal.api import ApiCaller, Packet

from . import Writer as WriteBase
from ..indicators import Indicator
from ..indicators.numeric import NumericIndicator

MAX_BUFFER_SIZE = 1024
WARMUP_COMMITS = 5


class Writer(WriteBase):
    def __init__(self, api_caller: ApiCaller, *,
                 frequency: float):
        super().__init__()

        self.frequency = frequency
        self.api_caller = api_caller
        self.last_committed = time.time()
        self.commits_count = 0
        self.indicators = {}
        self.api_caller.add_state_attribute('wildcard_indicators')
        self.api_caller.add_state_attribute('indicators')

    @staticmethod
    def _parse_key(key: str):
        return key

    def save_indicators(self, dot_indicators: Dict[str, Indicator], indicators: Dict[str, Indicator]):
        wildcards = {k: ind.to_dict() for k, ind in dot_indicators.items()}
        inds = {k: ind.to_dict() for k, ind in indicators.items()}

        self.api_caller.push(Packet({'wildcard_indicators': wildcards, 'indicators': inds}))

    def _write_indicator(self, global_step: int, indicator: Indicator):
        if indicator.is_empty():
            return

        if isinstance(indicator, NumericIndicator):
            value = indicator.get_mean()
            key = self._parse_key(indicator.mean_key)
            if key not in self.indicators:
                self.indicators[key] = []

            self.indicators[key].append((global_step, value))

    def write(self, *,
              global_step: int,
              indicators: Dict[str, Indicator]):
        for ind in indicators.values():
            self._write_indicator(global_step, ind)

        t = time.time()
        freq = self.frequency
        if self.commits_count < WARMUP_COMMITS:
            freq /= 2 ** (WARMUP_COMMITS - self.commits_count)
        if t - self.last_committed > freq:
            self.last_committed = t
            self.commits_count += 1
            self.flush()

    def flush(self):
        data = {}

        for key, value in self.indicators.items():
            value = np.array(value)
            step: np.ndarray = value[:, 0]
            value: np.ndarray = value[:, 1]
            while value.shape[0] > MAX_BUFFER_SIZE:
                if value.shape[0] % 2 == 1:
                    value = np.concatenate((value, value[-1:]))
                    step = np.concatenate((step, step[-1:]))

                n = value.shape[0] // 2
                step = np.mean(step.reshape(n, 2), axis=-1)
                value = np.mean(value.reshape(n, 2), axis=-1)

            data[key] = {
                'step': step.tolist(),
                'value': value.tolist()
            }

        self.indicators = {}

        self.api_caller.push(Packet({
            'track': data,
            'time': time.time()
        }))



from typing import Dict

try:
    import torch
except ImportError:
    torch = None


class Indicator:
    def __init__(self, *, name: str, is_print: bool):
        self.is_print = is_print
        self.name = name

    def clear(self):
        pass

    def is_empty(self) -> bool:
        raise NotImplementedError()

    def to_dict(self) -> Dict:
        return dict(class_name=self.__class__.__name__,
                    name=self.name,
                    is_print=self.is_print)

    def collect_value(self, value):
        raise NotImplementedError()

    def copy(self, key: str):
        raise NotImplementedError()

    def equals(self, value: any) -> bool:
        if type(value) != type(self):
            return False
        return value.name == self.name and value.is_print == self.is_print



from typing import Union, List, Tuple

from labml import logger
from labml.internal.util.colors import StyleCode
from labml.logger import Text


def labml_notice(message: Union[str, List[Union[str, Tuple[str, StyleCode]]]], *, is_danger=False, is_warn=True):
    log = [('\n' + '-' * 50, Text.subtle)]
    if is_danger:
        log.append(('\nLABML ERROR\n', [Text.danger, Text.title]))
    elif is_warn:
        log.append(('\nLABML WARNING\n', [Text.warning, Text.title]))
    else:
        log.append(('\nLABML MESSAGE\n', [Text.title]))

    if isinstance(message, str):
        log.append(message)
    else:
        log += message

    log.append(('\n' + '-' * 50, Text.subtle))
    logger.log(log)



from typing import List, Optional

import numpy as np
import tensorflow as tf
from tensorboard.backend.event_processing.directory_watcher import DirectoryDeletedError
from tensorboard.backend.event_processing.event_accumulator import EventAccumulator
from tensorboard.plugins.distribution import compressor

from .analytics import Analytics, BASIS_POINTS, Event


def _is_filtered(event, start_step: Optional[int], end_step: Optional[int]):
    if start_step is not None and event.step < start_step:
        return False
    elif end_step is not None and event.step >= end_step:
        return False
    else:
        return True


class TensorBoardAnalytics(Analytics):
    def __init__(self, log_path):
        self.event_acc = EventAccumulator(str(log_path), size_guidance={'tensors': 1000})

    def load(self):
        try:
            self.event_acc.Reload()
        except DirectoryDeletedError:
            raise FileNotFoundError()

    def tensor(self, name: str, start_step: Optional[int], end_step: Optional[int]) -> List[Event]:
        name = name.replace('.', '/')
        events = self.event_acc.Tensors(name)
        return [Event(e.step, tf.make_ndarray(e.tensor_proto))
                for e in events if _is_filtered(e, start_step, end_step)]

    def summarize(self, events: List[Event]):
        step = np.mean([e.step for e in events])
        values = np.sort([e.tensor for e in events])
        basis_points = np.percentile(values, BASIS_POINTS)

        return np.concatenate(([step], basis_points))

    def summarize_compressed_histogram(self, events: List[Event]):
        basis_points = [int(b) for b in np.multiply(BASIS_POINTS, 100)]
        results = []
        for e in events:
            buckets = compressor.compress_histogram(e.tensor)
            assert (len(buckets) == len(basis_points))
            for i, c in enumerate(buckets):
                assert (c.basis_point == basis_points[i])
            results.append([e.step] + [c.value for c in buckets])

        return np.asarray(results)



from typing import Tuple, Optional, List, overload, Union, TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    import torch

from labml.internal.analytics import cache as _cache
from labml.internal.analytics.altair import density as _density
from labml.internal.analytics.altair import histogram as _histogram
from labml.internal.analytics.altair import scatter as _scatter
from labml.internal.analytics.altair import binned_heatmap as _binned_heatmap
from labml.internal.analytics.indicators import IndicatorCollection as _IndicatorCollection


def _remove_names_prefix(names: List[Union[str, List[str]]]) -> List[str]:
    if len(names) == 0:
        return []

    if isinstance(names[0], list):
        common = names[0]
    else:
        common = None

    for n in names:
        if common is None:
            break
        if not isinstance(n, list):
            common = None
        merge = []
        for x, y in zip(common, n):
            if x != y:
                merge.append(None)
            else:
                merge.append(x)
        common = merge

    res = []
    for n in names:
        if isinstance(n, list):
            if common is not None:
                n = [p for i, p in enumerate(n) if i > len(common) or p != common[i]]
            n = '-'.join(n)

        res.append(n)

    return res


class IndicatorCollection(_IndicatorCollection):
    r"""
    You can get a indicator collection with :func:`runs`.

    >>> from labml import analytics
    >>> indicators = analytics.runs('1d3f855874d811eabb9359457a24edc8')

    You can reference individual indicators as attributes.

    >>> train_loss = indicators.train_loss

    You can add multiple indicator collections

    >>> losses = indicators.train_loss + indicators.validation_loss
    """
    pass


def runs(*uuids: str):
    r"""
    This is used to analyze runs.
    It fetches all the log indicators.

    Arguments:
        uuids (str): UUIDs of the runs. You can
            get this from `dashboard <https://github.com/lab-ml/lab_dashboard>`_

    Example:
        >>> from labml import analytics
        >>> indicators = analytics.runs('1d3f855874d811eabb9359457a24edc8')
    """

    indicators = None
    for r in uuids:
        run = _cache.get_run(r)
        indicators = indicators + run.indicators

    return indicators


def get_run(uuid: str):
    r"""
    Returns ``Run`` object
    """
    return _cache.get_run(uuid)


def set_preferred_db(db: str):
    if db not in ['tensorboard', 'sqlite']:
        raise ValueError('db should be one of tensorboard or sqlite')
    _cache.set_preferred_db(db)


@overload
def distribution(indicators: IndicatorCollection, *,
                 names: Optional[List[str]] = None,
                 levels: int = 5, alpha: int = 0.6,
                 height: int = 400, width: int = 800, height_minimap: int = 100):
    ...


@overload
def distribution(series: List[Union[np.ndarray, 'torch.Tensor']], *,
                 names: Optional[List[str]] = None,
                 levels: int = 5, alpha: int = 0.6,
                 height: int = 400, width: int = 800, height_minimap: int = 100):
    ...


@overload
def distribution(series: List[Union[np.ndarray, 'torch.Tensor']],
                 step: np.ndarray, *,
                 names: Optional[List[str]] = None,
                 levels: int = 5, alpha: int = 0.6,
                 height: int = 400, width: int = 800, height_minimap: int = 100):
    ...


@overload
def distribution(series: Union[np.ndarray, 'torch.Tensor'], *,
                 names: Optional[List[str]] = None,
                 levels: int = 5, alpha: int = 0.6,
                 height: int = 400, width: int = 800, height_minimap: int = 100):
    ...


def distribution(*args: any,
                 names: Optional[List[str]] = None,
                 levels: int = 5, alpha: int = 0.6,
                 height: int = 400, width: int = 800, height_minimap: int = 100):
    r"""
    Creates a distribution plot distribution with Altair

    This has multiple overloads

    .. function:: distribution(indicators: IndicatorCollection, *, names: Optional[List[str]] = None, levels: int = 5, alpha: int = 0.6, height: int = 400, width: int = 800, height_minimap: int = 100)
        :noindex:

    .. function:: distribution(series: Union[np.ndarray, torch.Tensor], *, names: Optional[List[str]] = None, levels: int = 5, alpha: int = 0.6, height: int = 400, width: int = 800, height_minimap: int = 100)
        :noindex:

    .. function:: distribution(series: List[Union[np.ndarray, torch.Tensor]], *, names: Optional[List[str]] = None, levels: int = 5, alpha: int = 0.6, height: int = 400, width: int = 800, height_minimap: int = 100)
        :noindex:

    .. function:: distribution(series: List[Union[np.ndarray, torch.Tensor]], step: np.ndarray, *, names: Optional[List[str]] = None, levels: int = 5, alpha: int = 0.6, height: int = 400, width: int = 800, height_minimap: int = 100)
        :noindex:

    Arguments:
        indicators(IndicatorCollection): Set of indicators to be plotted
        series(List[np.ndarray]): List of series of data
        step(np.ndarray): Steps

    Keyword Arguments:
        names(List[str]): List of names of series
        levels: how many levels of the distribution to be plotted
        alpha: opacity of the distribution
        height: height of the visualization
        width: width of the visualization
        height_minimap: height of the view finder

    Return:
        The Altair visualization

    Example:
        >>> from labml import analytics
        >>> indicators = analytics.runs('1d3f855874d811eabb9359457a24edc8')
        >>> analytics.distribution(indicators)
    """

    series = None
    step = None

    if len(args) == 1:
        if isinstance(args[0], _IndicatorCollection):
            series, names_ = _cache.get_indicators_data(args[0])
            if not series:
                raise ValueError("No series found")
            if names is None:
                names = names_
        elif isinstance(args[0], list):
            series = args[0]
        else:
            series = [args[0]]
    elif len(args) == 2:
        series = args[0]
        step = args[1]

    if names is None:
        names = [f'{i + 1}' for i in range(len(series))]

    if series is None:
        raise ValueError("distribution should be called with an indicator collection"
                         " or a series. Check documentation for details.")

    names = _remove_names_prefix(names)
    tables = _density.data_to_table(series, names, step)

    return _density.render(
        tables,
        levels=levels,
        alpha=alpha,
        width=width,
        height=height,
        height_minimap=height_minimap)


def histogram(series: Union[np.ndarray, 'torch.Tensor'], *,
              low: Optional[float] = None, high: Optional[float] = None,
              height: int = 400, width: int = 800, height_minimap: int = 100):
    r"""
    Creates a histogram with Altair

    Arguments:
        series(Union[np.ndarray, torch.Tensor]): Data

    Keyword Arguments:
        low: values less than this are ignored
        high: values greater than this are ignored
        height: height of the visualization
        width: width of the visualization
        height_minimap: height of the view finder

    Return:
        The Altair visualization
    """

    table = _histogram.data_to_table(series, low, high)

    return _histogram.render(
        table,
        width=width,
        height=height,
        height_minimap=height_minimap)


@overload
def scatter(indicators: IndicatorCollection, x_indicators: IndicatorCollection, *,
            names: Optional[List[str]] = None, x_name: Optional[str] = None,
            noise: Optional[Tuple[float, float]] = None,
            circle_size: int = 20,
            height: int = 400, width: int = 800, height_minimap: int = 100):
    ...


@overload
def scatter(series: List[np.ndarray],
            x_series: np.ndarray, *,
            names: Optional[List[str]] = None, x_name: Optional[str] = None,
            noise: Optional[Tuple[float, float]] = None,
            circle_size: int = 20,
            height: int = 400, width: int = 800, height_minimap: int = 100):
    ...


def scatter(*args: any,
            names: Optional[List[str]] = None, x_name: Optional[str] = None,
            noise: Optional[Tuple[float, float]] = None,
            circle_size: int = 20,
            height: int = 400, width: int = 800, height_minimap: int = 100):
    r"""
    Creates a scatter plot with Altair

    This has multiple overloads

    .. function:: scatter(indicators: IndicatorCollection, x_indicators: IndicatorCollection, *, names: Optional[List[str]] = None, x_name: Optional[str] = None, noise: Optional[Tuple[float, float]] = None, circle_size: int = 20, height: int = 400, width: int = 800, height_minimap: int = 100)
        :noindex:

    .. function:: scatter(series: List[np.ndarray], x_series: np.ndarray, *, names: Optional[List[str]] = None, x_name: Optional[str] = None, noise: Optional[Tuple[float, float]] = None, circle_size: int = 20, height: int = 400, width: int = 800, height_minimap: int = 100)
        :noindex:

    Arguments:
        indicators(IndicatorCollection): Set of indicators to be plotted
        x_indicators(IndicatorCollection): Indicator for x-axis
        series(List[np.ndarray]): List of series of data
        x_series(np.ndarray): X series of data

    Keyword Arguments:
        names(List[str]): List of names of series
        name(str): Name of X series
        noise: Noise to be added to spread out the scatter plot
        circle_size: size of circles in the plot
        height: height of the visualization
        width: width of the visualization
        height_minimap: height of the view finder

    Return:
        The Altair visualization

    :Example:
        >>> from labml import analytics
        >>> indicators = analytics.runs('1d3f855874d811eabb9359457a24edc8')
        >>> analytics.scatter(indicators.validation_loss, indicators.train_loss)
    """

    series = None
    x_series = None

    if len(args) == 2:
        if isinstance(args[0], _IndicatorCollection) and isinstance(args[1], _IndicatorCollection):
            series, names_ = _cache.get_indicators_data(args[0])
            x_series, x_name_ = _cache.get_indicators_data(args[1])

            if len(x_series) != 1:
                raise ValueError("There should be exactly one series for x-axis")
            if not series:
                raise ValueError("No series found")
            x_series = x_series[0]
            if x_name is None:
                x_name = x_name_[0]
            if names is None:
                names = names_
        elif isinstance(args[0], list):
            series = args[0]
            x_series = args[1]

    if series is None:
        raise ValueError("scatter should be called with an indicator collection"
                         " or a series. Check documentation for details.")

    if x_name is None:
        x_name = 'x'
    if names is None:
        names = [f'{i + 1}' for i in range(len(series))]

    tables = [_scatter.data_to_table(s, x_series, noise) for s in series]
    names = _remove_names_prefix(names)

    return _scatter.render(
        tables,
        names=names,
        x_name=x_name,
        width=width,
        height=height,
        height_minimap=height_minimap,
        circle_size=circle_size)


@overload
def binned_heatmap(indicators: IndicatorCollection, x_indicators: IndicatorCollection, *,
                   names: Optional[List[str]] = None, x_name: Optional[str] = None,
                   height: int = 400, width: int = 800, height_minimap: int = 100):
    ...


@overload
def binned_heatmap(series: List[np.ndarray],
                   x_series: np.ndarray, *,
                   names: Optional[List[str]] = None, x_name: Optional[str] = None,
                   height: int = 400, width: int = 800, height_minimap: int = 100):
    ...


def binned_heatmap(*args: any,
                   names: Optional[List[str]] = None, x_name: Optional[str] = None,
                   height: int = 400, width: int = 800, height_minimap: int = 100):
    r"""
    Creates a scatter plot with Altair

    This has multiple overloads

    .. function:: binned_heatmap(indicators: IndicatorCollection, x_indicators: IndicatorCollection, *, names: Optional[List[str]] = None, x_name: Optional[str] = None, height: int = 400, width: int = 800, height_minimap: int = 100)
        :noindex:

    .. function:: binned_heatmap(series: List[np.ndarray], x_series: np.ndarray, *, names: Optional[List[str]] = None, x_name: Optional[str] = None, height: int = 400, width: int = 800, height_minimap: int = 100)
        :noindex:

    Arguments:
        indicators(IndicatorCollection): Set of indicators to be plotted
        x_indicators(IndicatorCollection): Indicator for x-axis
        series(List[np.ndarray]): List of series of data
        x_series(np.ndarray): X series of data

    Keyword Arguments:
        names(List[str]): List of names of series
        name(str): Name of X series
        noise: Noise to be added to spread out the scatter plot
        circle_size: size of circles in the plot
        height: height of the visualization
        width: width of the visualization
        height_minimap: height of the view finder

    Return:
        The Altair visualization

    Example:
        >>> from labml import analytics
        >>> indicators = analytics.runs('1d3f855874d811eabb9359457a24edc8')
        >>> analytics.scatter(indicators.validation_loss, indicators.train_loss)
    """

    series = None
    x_series = None

    if len(args) == 2:
        if isinstance(args[0], _IndicatorCollection) and isinstance(args[1], _IndicatorCollection):
            series, names_ = _cache.get_indicators_data(args[0])
            x_series, x_name_ = _cache.get_indicators_data(args[1])

            if len(x_series) != 1:
                raise ValueError("There should be exactly one series for x-axis")
            if not series:
                raise ValueError("No series found")
            x_series = x_series[0]
            if x_name is None:
                x_name = x_name_[0]
            if names is None:
                names = names_
        elif isinstance(args[0], list):
            series = args[0]
            x_series = args[1]

    if series is None:
        raise ValueError("scatter should be called with an indicator collection"
                         " or a series. Check documentation for details.")

    if x_name is None:
        x_name = 'x'
    if names is None:
        names = [f'{i + 1}' for i in range(len(series))]

    tables = [_binned_heatmap.data_to_table(s, x_series) for s in series]
    names = _remove_names_prefix(names)

    return _binned_heatmap.render(
        tables,
        names=names,
        x_name=x_name,
        width=width,
        height=height,
        height_minimap=height_minimap)


def indicator_data(indicators: IndicatorCollection) -> Tuple[List[np.ndarray], List[List[str]]]:
    r"""
    Returns a tuple of a list of series and a list of names of series.
    Each series, `S` is a timeseries of histograms of shape `[T, 10]`,
    where `T` is the number of timesteps.
    `S[:, 0]` is the `global_step`.
    `S[:, 1:10]` represents the distribution at basis points
    `0, 6.68, 15.87, 30.85, 50.00, 69.15, 84.13, 93.32, 100.00`.

    Example:
        >>> from labml import analytics
        >>> indicators = analytics.runs('1d3f855874d811eabb9359457a24edc8')
        >>> analytics.indicator_data(indicators)
    """

    series, names = _cache.get_indicators_data(indicators)

    if not series:
        raise ValueError("No series found")

    return series, names


def artifact_data(indicators: IndicatorCollection) -> Tuple[List[any], List[str]]:
    r"""
    Returns a tuple of a list of series and a list of names of series.
    Each series, ``S`` is a timeseries of histograms of shape ``[T, 10]``,
    where ``T`` is the number of timesteps.
    ``S[:, 0]`` is the `global_step`.
    ``S[:, 1:10]`` represents the distribution at basis points:
    ``0, 6.68, 15.87, 30.85, 50.00, 69.15, 84.13, 93.32, 100.00``.

    Example:
        >>> from labml import analytics
        >>> indicators = analytics.runs('1d3f855874d811eabb9359457a24edc8')
        >>> analytics.artifact_data(indicators)
    """

    series, names = _cache.get_artifacts_data(indicators)

    if not series:
        raise ValueError("No series found")

    return series, names



from pathlib import PurePath
from typing import Dict

from ..indicators import Indicator
from ..indicators.artifacts import Image
from ..indicators.numeric import NumericIndicator

from . import Writer as WriteBase
from labml.internal.util.tensorboard_writer import get_tensorboard_writer


class Writer(WriteBase):
    def __init__(self, log_path: PurePath):
        super().__init__()

        self.__writer = get_tensorboard_writer(log_path)

    @staticmethod
    def _parse_key(key: str):
        return key.replace('.', '/')

    def write_h_parameters(self, hparams: Dict[str, any]):
        self.__writer.hparams(hparams)

    def _write_indicator(self, global_step: int, indicator: Indicator):
        if indicator.is_empty():
            return

        if isinstance(indicator, NumericIndicator):
            hist_data = indicator.get_histogram()
            if hist_data is not None:
                self.__writer.add_histogram(self._parse_key(indicator.name), hist_data, global_step)

            self.__writer.add_scalar(self._parse_key(indicator.mean_key), indicator.get_mean(), global_step)

        if isinstance(indicator, Image):
            for key in indicator.keys():
                self.__writer.add_image(self._parse_key(indicator.name), indicator.get_value(key), global_step)

    def write(self, *,
              global_step: int,
              indicators: Dict[str, Indicator]):
        for ind in indicators.values():
            self._write_indicator(global_step, ind)



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from . import Tracker


class Namespace:
    def __init__(self, *,
                 tracker: 'Trakcer',
                 name: str):
        self.tracker = tracker
        self.name = name

    def __enter__(self):
        self.tracker.namespace_enter(self)

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.tracker.namespace_exit(self)



from typing import TYPE_CHECKING, Type

if TYPE_CHECKING:
    from .base import Configs


class ConfigItem:
    def __init__(self, *,
                 key: str,
                 configs_class: Type['Configs'],
                 has_annotation: bool, annotation: any,
                 has_value: bool, value: any):
        self.key = key
        if annotation is None:
            annotation = type(value)
        self.annotation = annotation
        self.value = value
        self.has_annotation = has_annotation
        self.has_value = has_value
        self.configs_class = configs_class

    def update(self, k: 'ConfigItem'):
        if k.has_annotation:
            self.has_annotation = True
            self.annotation = k.annotation

        if k.has_value:
            self.has_value = True
            self.value = k.value



from typing import Optional, Dict, List, Callable, Any

import torch.optim
import torch.optim
import torch.utils.data
import torch.utils.data
from torch import nn

import labml.utils.pytorch as pytorch_utils
from labml import tracker, monit
from labml.configs import option, meta_config
from .device import DeviceConfigs
from .metrics import StateModule
from .training_loop import TrainingLoopConfigs


class ModeState:
    def __init__(self):
        self._rollback_stack = []

        self.is_train = False
        self.is_log_activations = False
        self.is_log_parameters = False
        self.is_optimize = False

    def _enter(self, mode: Dict[str, any]):
        rollback = {}
        for k, v in mode.items():
            if v is None:
                continue
            rollback[k] = getattr(self, k)
            setattr(self, k, v)

        self._rollback_stack.append(rollback)

        return len(self._rollback_stack)

    def _exit(self, n: int):
        assert n == len(self._rollback_stack)

        rollback = self._rollback_stack[-1]
        self._rollback_stack.pop(-1)

        for k, v in rollback.items():
            setattr(self, k, v)

    def update(self, *,
               is_train: Optional[bool] = None,
               is_log_parameters: Optional[bool] = None,
               is_log_activations: Optional[bool] = None,
               is_optimize: Optional[bool] = None):
        return Mode(self,
                    is_train=is_train,
                    is_log_parameters=is_log_parameters,
                    is_log_activations=is_log_activations,
                    is_optimize=is_optimize)


class Mode:
    def __init__(self, mode: ModeState, **kwargs: any):
        self.mode = mode
        self.update = {}
        for k, v in kwargs.items():
            if v is not None:
                self.update[k] = v

        self.idx = -1

    def __enter__(self):
        self.idx = self.mode._enter(self.update)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.mode._exit(self.idx)


class ForwardHook:
    def __init__(self, mode: ModeState, model_name, name: str, module: torch.nn.Module):
        self.mode = mode
        self.model_name = model_name
        self.name = name
        self.module = module
        module.register_forward_hook(self)

    def save(self, name: str, output):
        if isinstance(output, torch.Tensor):
            pytorch_utils.store_l1_l2(name, output)
        elif isinstance(output, tuple):
            for i, o in enumerate(output):
                self.save(f"{name}.{i}", o)

    def __call__(self, module, i, o):
        if not self.mode.is_log_activations:
            return

        self.save(f"module.{self.model_name}.{self.name}", o)


def hook_model_outputs(mode: ModeState, model: torch.nn.Module, model_name: str = "model"):
    for name, module in model.named_modules():
        if name == '':
            name = 'full'
        ForwardHook(mode, model_name, name, module)


class Trainer:
    def __init__(self, *,
                 name: str,
                 mode: ModeState,
                 data_loader: torch.utils.data.DataLoader,
                 inner_iterations: int,
                 state_modules: List[StateModule],
                 step: Callable[[any, 'BatchIndex'], None]):
        self.mode = mode
        self.name = name
        self.step = step
        self.data_loader = data_loader
        self.state_modules = state_modules
        self.__iterable = None
        self.__states = [sm.create_state() for sm in self.state_modules]
        self._batch_index = BatchIndex(len(data_loader), inner_iterations)

    def __call__(self):
        for sm, s in zip(self.state_modules, self.__states):
            sm.set_state(s)

        if self.__iterable is None or self._batch_index.completed:
            self.__iterable = iter(self.data_loader)
            self._batch_index.reset()
            for sm in self.state_modules:
                sm.on_epoch_start()
        with torch.set_grad_enabled(self.mode.is_train):
            self.__iterate()

        if self._batch_index.completed:
            for sm in self.state_modules:
                sm.on_epoch_end()

    def __iterate(self):
        with monit.section(self.name, is_partial=True):
            if self._batch_index.idx == 0:
                monit.progress(0)
            while not self._batch_index.iteration_completed:
                batch = next(self.__iterable)

                self.step(batch, self._batch_index)

                self._batch_index.step()
                monit.progress(self._batch_index.epoch_progress)

        self._batch_index.step_inner()


class BatchIndex:
    idx: int
    total: int
    iteration: int
    total_iterations: int

    def __init__(self, total: int, total_iterations: int):
        self.total_iterations = total_iterations
        self.total = total

    def is_interval(self, interval: int):
        if interval <= 0:
            return False
        if self.idx + 1 == self.total:
            return True
        else:
            return (self.idx + 1) % interval == 0

    @property
    def is_last(self):
        return self.idx + 1 == self.total

    @property
    def completed(self):
        return self.iteration >= self.total_iterations

    @property
    def iteration_completed(self):
        # // is important so that the last step happens on the last iteration
        return self.idx >= (self.iteration + 1) * self.total // self.total_iterations

    @property
    def epoch_progress(self):
        return self.idx / self.total

    def step(self):
        self.idx += 1

    def step_inner(self):
        self.iteration += 1

    def reset(self):
        self.idx = 0
        self.iteration = 0


class TrainValidConfigs(TrainingLoopConfigs):
    state_modules: List[StateModule]

    mode: ModeState

    epochs: int = 10

    trainer: Trainer
    validator: Trainer
    train_loader: torch.utils.data.DataLoader
    valid_loader: torch.utils.data.DataLoader

    loop_count = '_data_loop_count'
    loop_step = None

    inner_iterations: int = 1

    def init(self):
        pass

    def step(self, batch: Any, batch_idx: BatchIndex):
        raise NotImplementedError

    def run_step(self):
        for i in range(self.inner_iterations):
            with tracker.namespace('sample'):
                self.sample()
            with self.mode.update(is_train=True):
                with tracker.namespace('train'):
                    self.trainer()
            if self.validator:
                with tracker.namespace('valid'):
                    self.validator()
            tracker.save()

    def run(self):
        with monit.section("Initialize"):
            self.init()
        _ = self.validator
        _ = self.trainer
        for _ in self.training_loop:
            self.run_step()

    def sample(self):
        pass


@option(TrainValidConfigs.trainer)
def _default_trainer(c: TrainValidConfigs):
    return Trainer(name='Train',
                   mode=c.mode,
                   data_loader=c.train_loader,
                   inner_iterations=c.inner_iterations,
                   state_modules=c.state_modules,
                   step=c.step)


@option(TrainValidConfigs.validator)
def _default_validator(c: TrainValidConfigs):
    return Trainer(name='Valid',
                   mode=c.mode,
                   data_loader=c.valid_loader,
                   inner_iterations=c.inner_iterations,
                   state_modules=c.state_modules,
                   step=c.step)


@option(TrainValidConfigs.loop_count)
def _data_loop_count(c: TrainValidConfigs):
    return c.epochs


class SimpleTrainValidConfigs(TrainValidConfigs):
    optimizer: torch.optim.Adam
    model: nn.Module
    device: torch.device = DeviceConfigs()

    loss_func: nn.Module

    update_batches: int = 1
    log_params_updates: int = 2 ** 32  # 0 if not
    log_activations_batches: int = 2 ** 32  # 0 if not
    log_save_batches: int = 1

    state_modules: List[StateModule] = []

    def init(self):
        pass

    def step(self, batch: Any, batch_idx: BatchIndex):
        self.model.train(self.mode.is_train)
        data, target = batch[0].to(self.device), batch[1].to(self.device)

        if self.mode.is_train:
            tracker.add_global_step(len(data))

        is_log_activations = batch_idx.is_interval(self.log_activations_batches)
        with monit.section("model"):
            with self.mode.update(is_log_activations=is_log_activations):
                output = self.model(data)

        loss = self.loss_func(output, target)
        tracker.add("loss.", loss)

        if self.mode.is_train:
            with monit.section('backward'):
                loss.backward()

            if batch_idx.is_interval(self.update_batches):
                with monit.section('optimize'):
                    self.optimizer.step()
                if batch_idx.is_interval(self.log_params_updates):
                    pytorch_utils.store_model_indicators(self.model)
                self.optimizer.zero_grad()

            if batch_idx.is_interval(self.log_save_batches):
                tracker.save()


meta_config(SimpleTrainValidConfigs.update_batches,
            SimpleTrainValidConfigs.log_params_updates,
            SimpleTrainValidConfigs.log_activations_batches)


@option(SimpleTrainValidConfigs.optimizer)
def _default_optimizer(c: SimpleTrainValidConfigs):
    from labml_helpers.optimizer import OptimizerConfigs
    opt_conf = OptimizerConfigs()
    opt_conf.parameters = c.model.parameters()
    return opt_conf



"""
---
title: DQN Experiment with Atari Breakout
summary: Implementation of DQN experiment with Atari Breakout
---

# DQN Experiment with Atari Breakout

This experiment trains a Deep Q Network (DQN) to play Atari Breakout game on OpenAI Gym.
It runs the [game environments on multiple processes](../game.html) to sample efficiently.
"""

import numpy as np
import torch

from labml import tracker, experiment, logger, monit
from labml_helpers.schedule import Piecewise
from labml_nn.rl.dqn import QFuncLoss
from labml_nn.rl.dqn.model import Model
from labml_nn.rl.dqn.replay_buffer import ReplayBuffer
from labml_nn.rl.game import Worker

# Select device
if torch.cuda.is_available():
    device = torch.device("cuda:0")
else:
    device = torch.device("cpu")


def obs_to_torch(obs: np.ndarray) -> torch.Tensor:
    """Scale observations from `[0, 255]` to `[0, 1]`"""
    return torch.tensor(obs, dtype=torch.float32, device=device) / 255.


class Trainer:
    """
    ## Trainer
    """

    def __init__(self):
        # #### Configurations

        # number of workers
        self.n_workers = 8
        # steps sampled on each update
        self.worker_steps = 4
        # number of training iterations
        self.train_epochs = 8

        # number of updates
        self.updates = 1_000_000
        # size of mini batch for training
        self.mini_batch_size = 32

        # exploration as a function of updates
        self.exploration_coefficient = Piecewise(
            [
                (0, 1.0),
                (25_000, 0.1),
                (self.updates / 2, 0.01)
            ], outside_value=0.01)

        # update target network every 250 update
        self.update_target_model = 250

        # $\beta$ for replay buffer as a function of updates
        self.prioritized_replay_beta = Piecewise(
            [
                (0, 0.4),
                (self.updates, 1)
            ], outside_value=1)

        # Replay buffer with $\alpha = 0.6$. Capacity of the replay buffer must be a power of 2.
        self.replay_buffer = ReplayBuffer(2 ** 14, 0.6)

        # Model for sampling and training
        self.model = Model().to(device)
        # target model to get $\color{orange}Q(s';\color{orange}{\theta_i^{-}})$
        self.target_model = Model().to(device)

        # create workers
        self.workers = [Worker(47 + i) for i in range(self.n_workers)]

        # initialize tensors for observations
        self.obs = np.zeros((self.n_workers, 4, 84, 84), dtype=np.uint8)
        for worker in self.workers:
            worker.child.send(("reset", None))
        for i, worker in enumerate(self.workers):
            self.obs[i] = worker.child.recv()

        # loss function
        self.loss_func = QFuncLoss(0.99)
        # optimizer
        self.optimizer = torch.optim.Adam(self.model.parameters(), lr=2.5e-4)

    def _sample_action(self, q_value: torch.Tensor, exploration_coefficient: float):
        """
        #### $\epsilon$-greedy Sampling
        When sampling actions we use a $\epsilon$-greedy strategy, where we
        take a greedy action with probabiliy $1 - \epsilon$ and
        take a random action with probability $\epsilon$.
        We refer to $\epsilon$ as `exploration_coefficient`.
        """

        # Sampling doesn't need gradients
        with torch.no_grad():
            # Sample the action with highest Q-value. This is the greedy action.
            greedy_action = torch.argmax(q_value, dim=-1)
            # Uniformly sample and action
            random_action = torch.randint(q_value.shape[-1], greedy_action.shape, device=q_value.device)
            # Whether to chose greedy action or the random action
            is_choose_rand = torch.rand(greedy_action.shape, device=q_value.device) < exploration_coefficient
            # Pick the action based on `is_choose_rand`
            return torch.where(is_choose_rand, random_action, greedy_action).cpu().numpy()

    def sample(self, exploration_coefficient: float):
        """### Sample data"""

        # This doesn't need gradients
        with torch.no_grad():
            # Sample `worker_steps`
            for t in range(self.worker_steps):
                # Get Q_values for the current observation
                q_value = self.model(obs_to_torch(self.obs))
                # Sample actions
                actions = self._sample_action(q_value, exploration_coefficient)

                # Run sampled actions on each worker
                for w, worker in enumerate(self.workers):
                    worker.child.send(("step", actions[w]))

                # Collect information from each worker
                for w, worker in enumerate(self.workers):
                    # Get results after executing the actions
                    next_obs, reward, done, info = worker.child.recv()

                    # Add transition to replay buffer
                    self.replay_buffer.add(self.obs[w], actions[w], reward, next_obs, done)

                    # update episode information
                    # collect episode info, which is available if an episode finished;
                    #  this includes total reward and length of the episode -
                    #  look at `Game` to see how it works.
                    if info:
                        tracker.add('reward', info['reward'])
                        tracker.add('length', info['length'])

                    # update current observation
                    self.obs[w] = next_obs

    def train(self, beta: float):
        """
        ### Train the model
        """
        for _ in range(self.train_epochs):
            # Sample from priority replay buffer
            samples = self.replay_buffer.sample(self.mini_batch_size, beta)
            # Get the predicted Q-value
            q_value = self.model(obs_to_torch(samples['obs']))

            # Get the Q-values of the next state for [Double Q-learning](index.html).
            # Gradients shouldn't propagate for these
            with torch.no_grad():
                # Get $\color{cyan}Q(s';\color{cyan}{\theta_i})$
                double_q_value = self.model(obs_to_torch(samples['next_obs']))
                # Get $\color{orange}Q(s';\color{orange}{\theta_i^{-}})$
                target_q_value = self.target_model(obs_to_torch(samples['next_obs']))

            # Compute Temporal Difference (TD) errors, $\delta$, and the loss, $\mathcal{L}(\theta)$.
            td_errors, loss = self.loss_func(q_value,
                                             q_value.new_tensor(samples['action']),
                                             double_q_value, target_q_value,
                                             q_value.new_tensor(samples['done']),
                                             q_value.new_tensor(samples['reward']),
                                             q_value.new_tensor(samples['weights']))

            # Calculate priorities for replay buffer $p_i = |\delta_i| + \epsilon$
            new_priorities = np.abs(td_errors.cpu().numpy()) + 1e-6
            # Update replay buffer priorities
            self.replay_buffer.update_priorities(samples['indexes'], new_priorities)

            # Zero out the previously calculated gradients
            self.optimizer.zero_grad()
            # Calculate gradients
            loss.backward()
            # Clip gradients
            torch.nn.utils.clip_grad_norm_(self.model.parameters(), max_norm=0.5)
            # Update parameters based on gradients
            self.optimizer.step()

    def run_training_loop(self):
        """
        ### Run training loop
        """

        # Last 100 episode information
        tracker.set_queue('reward', 100, True)
        tracker.set_queue('length', 100, True)

        # Copy to target network initially
        self.target_model.load_state_dict(self.model.state_dict())

        for update in monit.loop(self.updates):
            # $\epsilon$, exploration fraction
            exploration = self.exploration_coefficient(update)
            tracker.add('exploration', exploration)
            # $\beta$ for prioritized replay
            beta = self.prioritized_replay_beta(update)
            tracker.add('beta', beta)

            # Sample with current policy
            self.sample(exploration)

            # Start training after the buffer is full
            if self.replay_buffer.is_full():
                # Train the model
                self.train(beta)

                # Periodically update target network
                if update % self.update_target_model == 0:
                    self.target_model.load_state_dict(self.model.state_dict())

            # Save tracked indicators.
            tracker.save()
            # Add a new line to the screen periodically
            if (update + 1) % 1_000 == 0:
                logger.log()

    def destroy(self):
        """
        ### Destroy
        Stop the workers
        """
        for worker in self.workers:
            worker.child.send(("close", None))


def main():
    # Create the experiment
    experiment.create(name='dqn')
    # Initialize the trainer
    m = Trainer()
    # Run and monitor the experiment
    with experiment.start():
        m.run_training_loop()
    # Stop the workers
    m.destroy()


# ## Run it
if __name__ == "__main__":
    main()



"""
---
title: Sketch RNN
summary: >
  This is an annotated PyTorch implementation of the Sketch RNN from paper A Neural Representation of Sketch Drawings.
  Sketch RNN is a sequence-to-sequence model that generates sketches of objects such as bicycles, cats, etc.
---

# Sketch RNN

This is an annotated implementation of the paper
[A Neural Representation of Sketch Drawings](https://arxiv.org/abs/1704.03477).

Sketch RNN is a sequence-to-sequence variational auto-encoder.
Both encoder and decoder are recurrent neural network models.
It learns to reconstruct stroke based simple drawings, by predicting
a series of strokes.
Decoder predicts each stroke as a mixture of Gaussian's.

### Getting data
Download data from [Quick, Draw! Dataset](https://github.com/googlecreativelab/quickdraw-dataset).
There is a link to download `npz` files in *Sketch-RNN QuickDraw Dataset* section of the readme.
Place the downloaded `npz` file(s) in `data/sketch` folder.
This code is configured to use `bicycle` dataset.
You can change this in configurations.

### Acknowledgements
Took help from [PyTorch Sketch RNN](https://github.com/alexis-jacq/Pytorch-Sketch-RNN) project by
[Alexis David Jacq](https://github.com/alexis-jacq)
"""

import math
from typing import Optional, Tuple, Any

import numpy as np
import torch
import torch.nn as nn
from matplotlib import pyplot as plt
from torch import optim
from torch.utils.data import Dataset, DataLoader

import einops
from labml import lab, experiment, tracker, monit
from labml_helpers.device import DeviceConfigs
from labml_helpers.module import Module
from labml_helpers.optimizer import OptimizerConfigs
from labml_helpers.train_valid import TrainValidConfigs, hook_model_outputs, BatchIndex


class StrokesDataset(Dataset):
    """
    ## Dataset

    This class load and pre-process the data.
    """

    def __init__(self, dataset: np.array, max_seq_length: int, scale: Optional[float] = None):
        """
        `dataset` is a list of numpy arrays of shape [seq_len, 3].
        It is a sequence of strokes, and each stroke is represented by
        3 integers.
        First two are the displacements along x and y ($\Delta x$, $\Delta y$)
        And the last integer represents the state of the pen - $1$ if it's touching
        the paper and $0$ otherwise.
        """

        data = []
        # We iterate through each of the sequences and filter
        for seq in dataset:
            # Filter if the length of the the sequence of strokes is within our range
            if 10 < len(seq) <= max_seq_length:
                # Clamp $\Delta x$, $\Delta y$ to $[-1000, 1000]$
                seq = np.minimum(seq, 1000)
                seq = np.maximum(seq, -1000)
                # Convert to a floating point array and add to `data`
                seq = np.array(seq, dtype=np.float32)
                data.append(seq)

        # We then calculate the scaling factor which is the
        # standard deviation of ($\Delta x$, $\Delta y$) combined.
        # Paper notes that the mean is not adjusted for simplicity,
        # since the mean is anyway close to $0$.
        if scale is None:
            scale = np.std(np.concatenate([np.ravel(s[:, 0:2]) for s in data]))
        self.scale = scale

        # Get the longest sequence length among all sequences
        longest_seq_len = max([len(seq) for seq in data])

        # We initialize PyTorch data array with two extra steps for start-of-sequence (sos)
        # and end-of-sequence (eos).
        # Each step is a vector $(\Delta x, \Delta y, p_1, p_2, p_3)$.
        # Only one of $p_1, p_2, p_3$ is $1$ and the others are $0$.
        # They represent *pen down*, *pen up* and *end-of-sequence* in that order.
        # $p_1$ is $1$ if the pen touches the paper in the next step.
        # $p_2$ is $1$ if the pen doesn't touch the paper in the next step.
        # $p_3$ is $1$ if it is the end of the drawing.
        self.data = torch.zeros(len(data), longest_seq_len + 2, 5, dtype=torch.float)
        # The mask array is needs only one extra-step since it is for the outputs of the
        # decoder, which takes in `data[:-1]` and predicts next step.
        self.mask = torch.zeros(len(data), longest_seq_len + 1)

        for i, seq in enumerate(data):
            seq = torch.from_numpy(seq)
            len_seq = len(seq)
            # Scale and set $\Delta x, \Delta y$
            self.data[i, 1:len_seq + 1, :2] = seq[:, :2] / scale
            # $p_1$
            self.data[i, 1:len_seq + 1, 2] = 1 - seq[:, 2]
            # $p_2$
            self.data[i, 1:len_seq + 1, 3] = seq[:, 2]
            # $p_3$
            self.data[i, len_seq + 1:, 4] = 1
            # Mask is on until end of sequence
            self.mask[i, :len_seq + 1] = 1

        # Start-of-sequence is $(0, 0, 1, 0, 0)
        self.data[:, 0, 2] = 1

    def __len__(self):
        """Size of the dataset"""
        return len(self.data)

    def __getitem__(self, idx: int):
        """Get a sample"""
        return self.data[idx], self.mask[idx]


class BivariateGaussianMixture:
    """
    ## Bi-variate Gaussian mixture

    The mixture is represented by $\Pi$ and
    $\mathcal{N}(\mu_{x}, \mu_{y}, \sigma_{x}, \sigma_{y}, \rho_{xy})$.
    This class adjust temperatures and creates the categorical and gaussian
    distributions from the parameters.
    """

    def __init__(self, pi_logits: torch.Tensor, mu_x: torch.Tensor, mu_y: torch.Tensor,
                 sigma_x: torch.Tensor, sigma_y: torch.Tensor, rho_xy: torch.Tensor):
        self.pi_logits = pi_logits
        self.mu_x = mu_x
        self.mu_y = mu_y
        self.sigma_x = sigma_x
        self.sigma_y = sigma_y
        self.rho_xy = rho_xy

    @property
    def n_distributions(self):
        """Number of distributions in the mixture, $M$"""
        return self.pi_logits.shape[-1]

    def set_temperature(self, temperature: float):
        """
        Adjust by temperature $\tau$
        """
        # $$\hat{\Pi_k} \leftarrow \frac{\hat{\Pi_k}}{\tau}$$
        self.pi_logits /= temperature
        # $$\sigma^2_x \leftarrow \sigma^2_x \tau$$
        self.sigma_x *= math.sqrt(temperature)
        # $$\sigma^2_y \leftarrow \sigma^2_y \tau$$
        self.sigma_y *= math.sqrt(temperature)

    def get_distribution(self):
        # Clamp $\sigma_x$, $\sigma_y$ and $\rho_{xy}$ to avoid getting `NaN`s
        sigma_x = torch.clamp_min(self.sigma_x, 1e-5)
        sigma_y = torch.clamp_min(self.sigma_y, 1e-5)
        rho_xy = torch.clamp(self.rho_xy, -1 + 1e-5, 1 - 1e-5)

        # Get means
        mean = torch.stack([self.mu_x, self.mu_y], -1)
        # Get covariance matrix
        cov = torch.stack([
            sigma_x * sigma_x, rho_xy * sigma_x * sigma_y,
            rho_xy * sigma_x * sigma_y, sigma_y * sigma_y
        ], -1)
        cov = cov.view(*sigma_y.shape, 2, 2)

        # Create bi-variate normal distribution.
        #
        # 📝 It would be efficient to `scale_tril` matrix as `[[a, 0], [b, c]]`
        # where
        # $$a = \sigma_x, b = \rho_{xy} \sigma_y, c = \sigma_y \sqrt{1 - \rho^2_{xy}}$$.
        # But for simplicity we use co-variance matrix.
        # [This is a good resource](https://www2.stat.duke.edu/courses/Spring12/sta104.1/Lectures/Lec22.pdf)
        # if you want to read up more about bi-variate distributions, their co-variance matrix,
        # and probability density function.
        multi_dist = torch.distributions.MultivariateNormal(mean, covariance_matrix=cov)

        # Create categorical distribution $\Pi$ from logits
        cat_dist = torch.distributions.Categorical(logits=self.pi_logits)

        #
        return cat_dist, multi_dist


class EncoderRNN(Module):
    """
    ## Encoder module

    This consists of a bidirectional LSTM
    """

    def __init__(self, d_z: int, enc_hidden_size: int):
        super().__init__()
        # Create a bidirectional LSTM takes a sequence of
        # $(\Delta x, \Delta y, p_1, p_2, p_3)$ as input.
        self.lstm = nn.LSTM(5, enc_hidden_size, bidirectional=True)
        # Head to get $\mu$
        self.mu_head = nn.Linear(2 * enc_hidden_size, d_z)
        # Head to get $\hat{\sigma}$
        self.sigma_head = nn.Linear(2 * enc_hidden_size, d_z)

    def __call__(self, inputs: torch.Tensor, state=None):
        # The hidden state of the bidirectional LSTM is the concatenation of the
        # output of the last token in the forward direction and
        # and first token in the reverse direction.
        # Which is what we want.
        # $$h_{\rightarrow} = encode_{\rightarrow}(S),
        # h_{\leftarrow} = encode←_{\leftarrow}(S_{reverse}),
        # h = [h_{\rightarrow}; h_{\leftarrow}]$$
        _, (hidden, cell) = self.lstm(inputs.float(), state)
        # The state has shape `[2, batch_size, hidden_size]`
        # where the first dimension is the direction.
        # We rearrange it to get $h = [h_{\rightarrow}; h_{\leftarrow}]$
        hidden = einops.rearrange(hidden, 'fb b h -> b (fb h)')

        # $\mu$
        mu = self.mu_head(hidden)
        # $\hat{\sigma}$
        sigma_hat = self.sigma_head(hidden)
        # $\sigma = \exp(\frac{\hat{\sigma}}{2})$
        sigma = torch.exp(sigma_hat / 2.)

        # Sample $z = \mu + \sigma \cdot \mathcal{N}(0, I)$
        z = mu + sigma * torch.normal(mu.new_zeros(mu.shape), mu.new_ones(mu.shape))

        #
        return z, mu, sigma_hat


class DecoderRNN(Module):
    """
    ## Decoder module

    This consists of a LSTM
    """

    def __init__(self, d_z: int, dec_hidden_size: int, n_distributions: int):
        super().__init__()
        # LSTM takes $[(\Delta x, \Delta y, p_1, p_2, p_3); z]$ as input
        self.lstm = nn.LSTM(d_z + 5, dec_hidden_size)

        # Initial state of the LSTM is $[h_0; c_0] = \tanh(W_{z}z + b_z)$.
        # `init_state` is the linear transformation for this
        self.init_state = nn.Linear(d_z, 2 * dec_hidden_size)

        # This layer produces outputs for each of of the `n_distributions`.
        # Each distribution needs six parameters
        # $(\hat{\Pi_i}, \mu_{x_i}, \mu_{y_i}, \hat{\sigma_{x_i}}, \hat{\sigma_{y_i}} \hat{\rho_{xy_i}})$
        self.mixtures = nn.Linear(dec_hidden_size, 6 * n_distributions)

        # This head is for the logits $(\hat{q_1}, \hat{q_2}, \hat{q_3})$
        self.q_head = nn.Linear(dec_hidden_size, 3)
        # This is to calculate $\log(q_k)$ where
        # $$q_k = \operatorname{softmax}(\hat{q})_k = \frac{\exp(\hat{q_k})}{\sum_{j = 1}^3 \exp(\hat{q_j})}$$
        self.q_log_softmax = nn.LogSoftmax(-1)

        # These parameters are stored for future reference
        self.n_distributions = n_distributions
        self.dec_hidden_size = dec_hidden_size

    def __call__(self, x: torch.Tensor, z: torch.Tensor, state: Optional[Tuple[torch.Tensor, torch.Tensor]]):
        # Calculate the initial state
        if state is None:
            # $[h_0; c_0] = \tanh(W_{z}z + b_z)$
            h, c = torch.split(torch.tanh(self.init_state(z)), self.dec_hidden_size, 1)
            # `h` and `c` have shapes `[batch_size, lstm_size]`. We want to make them
            # to shape `[1, batch_size, lstm_size]` because that's the shape used in LSTM.
            state = (h.unsqueeze(0).contiguous(), c.unsqueeze(0).contiguous())

        # Run the LSTM
        outputs, state = self.lstm(x, state)

        # Get $\log(q)$
        q_logits = self.q_log_softmax(self.q_head(outputs))

        # Get $(\hat{\Pi_i}, \mu_{x,i}, \mu_{y,i}, \hat{\sigma_{x,i}},
        # \hat{\sigma_{y,i}} \hat{\rho_{xy,i}})$.
        # `torch.split` splits the output into 6 tensors of size `self.n_distribution`
        # across dimension `2`.
        pi_logits, mu_x, mu_y, sigma_x, sigma_y, rho_xy = \
            torch.split(self.mixtures(outputs), self.n_distributions, 2)

        # Create a bi-variate gaussian mixture
        # $\Pi$ and 
        # $\mathcal{N}(\mu_{x}, \mu_{y}, \sigma_{x}, \sigma_{y}, \rho_{xy})$
        # where
        # $$\sigma_{x,i} = \exp(\hat{\sigma_{x,i}}), \sigma_{y,i} = \exp(\hat{\sigma_{y,i}}),
        # \rho_{xy,i} = \tanh(\hat{\rho_{xy,i}})$$
        # and
        # $$\Pi_i = \operatorname{softmax}(\hat{\Pi})_i = \frac{\exp(\hat{\Pi_i})}{\sum_{j = 1}^3 \exp(\hat{\Pi_j})}$$
        #
        # $\Pi$ is the categorical probabilities of choosing the distribution out of the mixture
        # $\mathcal{N}(\mu_{x}, \mu_{y}, \sigma_{x}, \sigma_{y}, \rho_{xy})$.
        dist = BivariateGaussianMixture(pi_logits, mu_x, mu_y,
                                        torch.exp(sigma_x), torch.exp(sigma_y), torch.tanh(rho_xy))

        #
        return dist, q_logits, state


class ReconstructionLoss(Module):
    """
    ## Reconstruction Loss
    """

    def __call__(self, mask: torch.Tensor, target: torch.Tensor,
                 dist: 'BivariateGaussianMixture', q_logits: torch.Tensor):
        # Get $\Pi$ and $\mathcal{N}(\mu_{x}, \mu_{y}, \sigma_{x}, \sigma_{y}, \rho_{xy})$
        pi, mix = dist.get_distribution()
        # `target` has shape `[seq_len, batch_size, 5]` where the last dimension is the features
        # $(\Delta x, \Delta y, p_1, p_2, p_3)$.
        # We want to get $\Delta x, \Delta$ and get the probabilities from each of the distributions
        # in the mixture $\mathcal{N}(\mu_{x}, \mu_{y}, \sigma_{x}, \sigma_{y}, \rho_{xy})$.
        #
        # `xy` will have shape `[seq_len, batch_size, n_distributions, 2]`
        xy = target[:, :, 0:2].unsqueeze(-2).expand(-1, -1, dist.n_distributions, -1)
        # Calculate the probabilities
        # $$p(\Delta x, \Delta y) =
        # \sum_{j=1}^M \Pi_j \mathcal{N} \big( \Delta x, \Delta y \vert
        # \mu_{x,j}, \mu_{y,j}, \sigma_{x,j}, \sigma_{y,j}, \rho_{xy,j}
        # \big)$$
        probs = torch.sum(pi.probs * torch.exp(mix.log_prob(xy)), 2)

        # $$L_s = - \frac{1}{N_{max}} \sum_{i=1}^{N_s} \log \big (p(\Delta x, \Delta y) \big)$$
        # Although `probs` has $N_{max}$ (`longest_seq_len`) elements the sum is only taken
        # upto $N_s$ because the rest are masked out.
        #
        # It might feel like we should be taking the sum and dividing by $N_s$ and not $N_{max}$,
        # but this will give higher weight for individual predictions in shorter sequences.
        # We give equal weight to each prediction $p(\Delta x, \Delta y)$ when we divide by $N_{max}$
        loss_stroke = -torch.mean(mask * torch.log(1e-5 + probs))

        # $$L_p = - \frac{1}{N_{max}} \sum_{i=1}^{N_{max}} \sum_{k=1}^{3} p_{k,i} \log(q_{k,i})$$
        loss_pen = -torch.mean(target[:, :, 2:] * q_logits)

        # $$L_R = L_s + L_p$$
        return loss_stroke + loss_pen


class KLDivLoss(Module):
    """
    ## KL-Divergence loss

    This calculates the KL divergence between a given normal distribution and $\mathcal{N}(0, 1)$
    """

    def __call__(self, sigma_hat: torch.Tensor, mu: torch.Tensor):
        # $$L_{KL} = - \frac{1}{2 N_z} \bigg( 1 + \hat{\sigma} - \mu^2 - \exp(\hat{\sigma}) \bigg)$$
        return -0.5 * torch.mean(1 + sigma_hat - mu ** 2 - torch.exp(sigma_hat))


class Sampler:
    """
    ## Sampler

    This samples a sketch from the decoder and plots it
    """

    def __init__(self, encoder: EncoderRNN, decoder: DecoderRNN):
        self.decoder = decoder
        self.encoder = encoder

    def sample(self, data: torch.Tensor, temperature: float):
        # $N_{max}$
        longest_seq_len = len(data)

        # Get $z$ from the encoder
        z, _, _ = self.encoder(data)

        # Start-of-sequence stroke is $(0, 0, 1, 0, 0)$
        s = data.new_tensor([0, 0, 1, 0, 0])
        seq = [s]
        # Initial decoder is `None`.
        # The decoder will initialize it to $[h_0; c_0] = \tanh(W_{z}z + b_z)$
        state = None

        # We don't need gradients
        with torch.no_grad():
            # Sample $N_{max}$ strokes
            for i in range(longest_seq_len):
                # $[(\Delta x, \Delta y, p_1, p_2, p_3); z] is the input to the decoder$
                data = torch.cat([s.view(1, 1, -1), z.unsqueeze(0)], 2)
                # Get $\Pi$, $\mathcal{N}(\mu_{x}, \mu_{y}, \sigma_{x}, \sigma_{y}, \rho_{xy})$,
                # $q$ and the next state from the decoder
                dist, q_logits, state = self.decoder(data, z, state)
                # Sample a stroke
                s = self._sample_step(dist, q_logits, temperature)
                # Add the new stroke to the sequence of strokes
                seq.append(s)
                # Stop sampling if $p_3 = 1$. This indicates that sketching has stopped
                if s[4] == 1:
                    break

        # Create a PyTorch tensor of the sequence of strokes
        seq = torch.stack(seq)

        # Plot the sequence of strokes
        self.plot(seq)

    @staticmethod
    def _sample_step(dist: 'BivariateGaussianMixture', q_logits: torch.Tensor, temperature: float):
        # Set temperature $\tau$ for sampling. This is implemented in class `BivariateGaussianMixture`.
        dist.set_temperature(temperature)
        # Get temperature adjusted $\Pi$ and $\mathcal{N}(\mu_{x}, \mu_{y}, \sigma_{x}, \sigma_{y}, \rho_{xy})$
        pi, mix = dist.get_distribution()
        # Sample from $\Pi$ the index of the distribution to use from the mixture
        idx = pi.sample()[0, 0]

        # Create categorical distribution $q$ with log-probabilities `q_logits` or $\hat{q}$
        q = torch.distributions.Categorical(logits=q_logits / temperature)
        # Sample from $q$
        q_idx = q.sample()[0, 0]

        # Sample from the normal distributions in the mixture and pick the one indexed by `idx`
        xy = mix.sample()[0, 0, idx]

        # Create an empty stroke $(\Delta x, \Delta y, q_1, q_2, q_3)$
        stroke = q_logits.new_zeros(5)
        # Set $\Delta x, \Delta y$
        stroke[:2] = xy
        # Set $q_1, q_2, q_3$
        stroke[q_idx + 2] = 1
        #
        return stroke

    @staticmethod
    def plot(seq: torch.Tensor):
        # Take the cumulative sums of $(\Delta x, \Delta y)$ to get $$x, y)$
        seq[:, 0:2] = torch.cumsum(seq[:, 0:2], dim=0)
        # Create a new numpy array of the form $(x, y, q_2)$
        seq[:, 2] = seq[:, 3]
        seq = seq[:, 0:3].detach().cpu().numpy()

        # Split the array at points where $q_2$ is $1$.
        # That is split the array of strokes at the points where the pen is lifted from the paper.
        # This gives a list of sequence of strokes.
        strokes = np.split(seq, np.where(seq[:, 2] > 0)[0] + 1)
        # Plot each sequence of strokes
        for s in strokes:
            plt.plot(s[:, 0], -s[:, 1])
        # Don't show axes
        plt.axis('off')
        # Show the plot
        plt.show()


class Configs(TrainValidConfigs):
    """
    ## Configurations

    These are default configurations which can be later adjusted by passing a `dict`.
    """

    # Device configurations to pick the device to run the experiment
    device: torch.device = DeviceConfigs()
    #
    encoder: EncoderRNN
    decoder: DecoderRNN
    optimizer: optim.Adam
    sampler: Sampler

    dataset_name: str
    train_loader: DataLoader
    valid_loader: DataLoader
    train_dataset: StrokesDataset
    valid_dataset: StrokesDataset

    # Encoder and decoder sizes
    enc_hidden_size = 256
    dec_hidden_size = 512

    # Batch size
    batch_size = 100

    # Number of features in $z$
    d_z = 128
    # Number of distributions in the mixture, $M$
    n_distributions = 20

    # Weight of KL divergence loss, $w_{KL}$
    kl_div_loss_weight = 0.5
    # Gradient clipping
    grad_clip = 1.
    # Temperature $\tau$ for sampling
    temperature = 0.4

    # Filter out stroke sequences longer than $200$
    max_seq_length = 200

    epochs = 100

    kl_div_loss = KLDivLoss()
    reconstruction_loss = ReconstructionLoss()

    def init(self):
        # Initialize encoder & decoder
        self.encoder = EncoderRNN(self.d_z, self.enc_hidden_size).to(self.device)
        self.decoder = DecoderRNN(self.d_z, self.dec_hidden_size, self.n_distributions).to(self.device)

        # Set optimizer, things like type of optimizer and learning rate are configurable
        optimizer = OptimizerConfigs()
        optimizer.parameters = list(self.encoder.parameters()) + list(self.decoder.parameters())
        self.optimizer = optimizer

        # Create sampler
        self.sampler = Sampler(self.encoder, self.decoder)

        # `npz` file path is `data/sketch/[DATASET NAME].npz`
        path = lab.get_data_path() / 'sketch' / f'{self.dataset_name}.npz'
        # Load the numpy file.
        dataset = np.load(str(path), encoding='latin1', allow_pickle=True)

        # Create training dataset
        self.train_dataset = StrokesDataset(dataset['train'], self.max_seq_length)
        # Create validation dataset
        self.valid_dataset = StrokesDataset(dataset['valid'], self.max_seq_length, self.train_dataset.scale)

        # Create training data loader
        self.train_loader = DataLoader(self.train_dataset, self.batch_size, shuffle=True)
        # Create validation data loader
        self.valid_loader = DataLoader(self.valid_dataset, self.batch_size)

        # Add hooks to monitor layer outputs on Tensorboard
        hook_model_outputs(self.mode, self.encoder, 'encoder')
        hook_model_outputs(self.mode, self.decoder, 'decoder')

        # Configure the tracker to print the total train/validation loss
        tracker.set_scalar("loss.total.*", True)

        self.state_modules = []

    def step(self, batch: Any, batch_idx: BatchIndex):
        self.encoder.train(self.mode.is_train)
        self.decoder.train(self.mode.is_train)

        # Move `data` and `mask` to device and swap the sequence and batch dimensions.
        # `data` will have shape `[seq_len, batch_size, 5]` and
        # `mask` will have shape `[seq_len, batch_size]`.
        data = batch[0].to(self.device).transpose(0, 1)
        mask = batch[1].to(self.device).transpose(0, 1)

        # Increment step in training mode
        if self.mode.is_train:
            tracker.add_global_step(len(data))

        # Encode the sequence of strokes
        with monit.section("encoder"):
            # Get $z$, $\mu$, and $\hat{\sigma}$
            z, mu, sigma_hat = self.encoder(data)

        # Decode the mixture of distributions and $\hat{q}$
        with monit.section("decoder"):
            # Concatenate $[(\Delta x, \Delta y, p_1, p_2, p_3); z]$
            z_stack = z.unsqueeze(0).expand(data.shape[0] - 1, -1, -1)
            inputs = torch.cat([data[:-1], z_stack], 2)
            # Get mixture of distributions and $\hat{q}$
            dist, q_logits, _ = self.decoder(inputs, z, None)

        # Compute the loss
        with monit.section('loss'):
            # $L_{KL}$
            kl_loss = self.kl_div_loss(sigma_hat, mu)
            # $L_R$
            reconstruction_loss = self.reconstruction_loss(mask, data[1:], dist, q_logits)
            # $Loss = L_R + w_{KL} L_{KL}$
            loss = reconstruction_loss + self.kl_div_loss_weight * kl_loss

            # Track losses
            tracker.add("loss.kl.", kl_loss)
            tracker.add("loss.reconstruction.", reconstruction_loss)
            tracker.add("loss.total.", loss)

        # Only if we are in training state
        if self.mode.is_train:
            # Run optimizer
            with monit.section('optimize'):
                # Set `grad` to zero
                self.optimizer.zero_grad()
                # Compute gradients
                loss.backward()
                # Log model parameters and gradients
                if batch_idx.is_last:
                    tracker.add(encoder=self.encoder, decoder=self.decoder)
                # Clip gradients
                nn.utils.clip_grad_norm_(self.encoder.parameters(), self.grad_clip)
                nn.utils.clip_grad_norm_(self.decoder.parameters(), self.grad_clip)
                # Optimize
                self.optimizer.step()

        tracker.save()

    def sample(self):
        # Randomly pick a sample from validation dataset to encoder
        data, *_ = self.valid_dataset[np.random.choice(len(self.valid_dataset))]
        # Add batch dimension and move it to device
        data = data.unsqueeze(1).to(self.device)
        # Sample
        self.sampler.sample(data, self.temperature)


def main():
    configs = Configs()
    experiment.create(name="sketch_rnn")

    # Pass a dictionary of configurations
    experiment.configs(configs, {
        'optimizer.optimizer': 'Adam',
        # We use a learning rate of `1e-3` because we can see results faster.
        # Paper had suggested `1e-4`.
        'optimizer.learning_rate': 1e-3,
        # Name of the dataset
        'dataset_name': 'bicycle',
        # Number of inner iterations within an epoch to switch between training, validation and sampling.
        'inner_iterations': 10
    })

    with experiment.start():
        # Run the experiment
        configs.run()


if __name__ == "__main__":
    main()



"""
---
title: Generative Adversarial Networks (GAN)
summary: A simple PyTorch implementation/tutorial of Generative Adversarial Networks (GAN) loss functions.
---

# Generative Adversarial Networks (GAN)

This is an implementation of
[Generative Adversarial Networks](https://arxiv.org/abs/1406.2661).

The generator, $G(\pmb{z}; \theta_g)$ generates samples that match the
distribution of data, while the discriminator, $D(\pmb{x}; \theta_g)$
gives the probability that $\pmb{x}$ came from data rather than $G$.

We train $D$ and $G$ simultaneously on a two-player min-max game with value
function $V(G, D)$.

$$\min_G \max_D V(D, G) =
    \mathop{\mathbb{E}}_{\pmb{x} \sim p_{data}(\pmb{x})}
        \big[\log D(\pmb{x})\big] +
    \mathop{\mathbb{E}}_{\pmb{z} \sim p_{\pmb{z}}(\pmb{z})}
        \big[\log (1 - D(G(\pmb{z}))\big]
$$

$p_{data}(\pmb{x})$ is the probability distribution over data,
whilst $p_{\pmb{z}}(\pmb{z})$ probability distribution of $\pmb{z}$, which is set to
gaussian noise.

This file defines the loss functions. [Here](simple_mnist_experiment.html) is an MNIST example
with two multilayer perceptron for the generator and discriminator.
"""

import torch
import torch.nn as nn
import torch.utils.data
import torch.utils.data

from labml_helpers.module import Module


class DiscriminatorLogitsLoss(Module):
    """
    ## Discriminator Loss

    Discriminator should **ascend** on the gradient,

    $$\nabla_{\theta_d} \frac{1}{m} \sum_{i=1}^m \Bigg[
        \log D\Big(\pmb{x}^{(i)}\Big) +
        \log \Big(1 - D\Big(G\Big(\pmb{z}^{(i)}\Big)\Big)\Big)
    \Bigg]$$

    $m$ is the mini-batch size and $(i)$ is used to index samples in the mini-batch.
    $\pmb{x}$ are samples from $p_{data}$ and $\pmb{z}$ are samples from $p_z$.
    """

    def __init__(self, smoothing: float = 0.2):
        super().__init__()
        # We use PyTorch Binary Cross Entropy Loss, which is
        # $-\sum\Big[y \log(\hat{y}) + (1 - y) \log(1 - \hat{y})\Big]$,
        # where $y$ are the labels and $\hat{y}$ are the predictions.
        # *Note the negative sign*.
        # We use labels equal to $1$ for $\pmb{x}$ from $p_{data}$
        # and labels equal to $0$ for $\pmb{x}$ from $p_{G}.$
        # Then descending on the sum of these is same as ascending on
        # the above gradient.
        #
        # `BCEWithLogitsLoss` combines softmax and binary cross entropy loss.
        self.loss_true = nn.BCEWithLogitsLoss()
        self.loss_false = nn.BCEWithLogitsLoss()

        # We use label smoothing because it seems to work better in some cases
        self.smoothing = smoothing

        # Labels are registered as buffered and persistence is set to `False`.
        self.register_buffer('labels_true', _create_labels(256, 1.0 - smoothing, 1.0), False)
        self.register_buffer('labels_false', _create_labels(256, 0.0, smoothing), False)

    def __call__(self, logits_true: torch.Tensor, logits_false: torch.Tensor):
        """
        `logits_true` are logits from $D(\pmb{x}^{(i)})$ and
        `logits_false` are logits from $D(G(\pmb{z}^{(i)}))$
        """
        if len(logits_true) > len(self.labels_true):
            self.register_buffer("labels_true",
                                 _create_labels(len(logits_true), 1.0 - self.smoothing, 1.0, logits_true.device), False)
        if len(logits_false) > len(self.labels_false):
            self.register_buffer("labels_false",
                                 _create_labels(len(logits_false), 0.0, self.smoothing, logits_false.device), False)

        return (self.loss_true(logits_true, self.labels_true[:len(logits_true)]),
                self.loss_false(logits_false, self.labels_false[:len(logits_false)]))


class GeneratorLogitsLoss(Module):
    """
    ## Generator Loss

    Generator should **descend** on the gradient,

    $$\nabla_{\theta_g} \frac{1}{m} \sum_{i=1}^m \Bigg[
        \log \Big(1 - D\Big(G\Big(\pmb{z}^{(i)}\Big)\Big)\Big)
    \Bigg]$$
    """
    def __init__(self, smoothing: float = 0.2):
        super().__init__()
        self.loss_true = nn.BCEWithLogitsLoss()
        self.smoothing = smoothing
        # We use labels equal to $1$ for $\pmb{x}$ from $p_{G}.$
        # Then descending on this loss is same as descending on
        # the above gradient.
        self.register_buffer('fake_labels', _create_labels(256, 1.0 - smoothing, 1.0), False)

    def __call__(self, logits: torch.Tensor):
        if len(logits) > len(self.fake_labels):
            self.register_buffer("fake_labels",
                                 _create_labels(len(logits), 1.0 - self.smoothing, 1.0, logits.device), False)

        return self.loss_true(logits, self.fake_labels[:len(logits)])


def _create_labels(n: int, r1: float, r2: float, device: torch.device = None):
    """
    Create smoothed labels
    """
    return torch.empty(n, 1, requires_grad=False, device=device).uniform_(r1, r2)



from torch.utils.data import DataLoader
from torchvision import datasets, transforms

from labml import lab
from labml.configs import BaseConfigs, aggregate, option


def _dataset(is_train, transform):
    return datasets.MNIST(str(lab.get_data_path()),
                          train=is_train,
                          download=True,
                          transform=transform)


class MNISTConfigs(BaseConfigs):
    dataset_name: str = 'MNIST'
    dataset_transforms: transforms.Compose
    train_dataset: datasets.MNIST
    valid_dataset: datasets.MNIST

    train_loader: DataLoader
    valid_loader: DataLoader

    train_batch_size: int = 64
    valid_batch_size: int = 1024

    train_loader_shuffle: bool = True
    valid_loader_shuffle: bool = False


@option(MNISTConfigs.dataset_transforms)
def mnist_transforms():
    return transforms.Compose([
        transforms.ToTensor(),
        transforms.Normalize((0.1307,), (0.3081,))
    ])


@option(MNISTConfigs.train_dataset)
def mnist_train_dataset(c: MNISTConfigs):
    return _dataset(True, c.dataset_transforms)


@option(MNISTConfigs.valid_dataset)
def mnist_valid_dataset(c: MNISTConfigs):
    return _dataset(False, c.dataset_transforms)


@option(MNISTConfigs.train_loader)
def mnist_train_loader(c: MNISTConfigs):
    return DataLoader(c.train_dataset,
                      batch_size=c.train_batch_size,
                      shuffle=c.train_loader_shuffle)


@option(MNISTConfigs.valid_loader)
def mnist_valid_loader(c: MNISTConfigs):
    return DataLoader(c.valid_dataset,
                      batch_size=c.valid_batch_size,
                      shuffle=c.valid_loader_shuffle)


aggregate(MNISTConfigs.dataset_name, 'MNIST',
          (MNISTConfigs.dataset_transforms, 'mnist_transforms'),
          (MNISTConfigs.train_dataset, 'mnist_train_dataset'),
          (MNISTConfigs.valid_dataset, 'mnist_valid_dataset'),
          (MNISTConfigs.train_loader, 'mnist_train_loader'),
          (MNISTConfigs.valid_loader, 'mnist_valid_loader'))



from typing import TYPE_CHECKING, Union, Tuple, Dict

from labml.utils.pytorch import store_model_indicators, store_optimizer_indicators

from . import Indicator

if TYPE_CHECKING:
    from torch.nn import Module
    from torch.optim.optimizer import Optimizer


class PyTorchModule(Indicator):
    def __init__(self, name: str, is_print=False):
        super().__init__(name=name, is_print=is_print)

    def is_empty(self) -> bool:
        return True

    def collect_value(self, model: 'Module'):
        store_model_indicators(model, model_name=self.name)

    def get_histogram(self):
        return None

    def copy(self, key: str):
        return PyTorchModule(key, is_print=self.is_print)


class PyTorchOptimizer(Indicator):
    def __init__(self, name: str, is_print=False):
        super().__init__(name=name, is_print=is_print)

    def is_empty(self) -> bool:
        return True

    def collect_value(self, optimizer: Union['Optimizer', Tuple['Optimizer', Dict[str, 'Module']]]):
        if isinstance(optimizer, tuple):
            store_optimizer_indicators(optimizer[0], models=optimizer[1], optimizer_name=self.name)
        else:
            store_optimizer_indicators(optimizer, optimizer_name=self.name)

    def get_histogram(self):
        return None

    def copy(self, key: str):
        return PyTorchModule(key, is_print=self.is_print)



from typing import List, Union, Tuple, Optional

from labml.internal.logger.destinations import Destination
from labml.internal.util.colors import StyleCode, ANSI_RESET


class ConsoleDestination(Destination):
    @staticmethod
    def __ansi_code(text: str, color: List[StyleCode] or StyleCode or None):
        """
        ### Add ansi color codes
        """
        if color is None:
            return text
        elif isinstance(color, list):
            return "".join([c.ansi() for c in color]) + f"{text}{ANSI_RESET}"
        else:
            return f"{color.ansi()}{text}{ANSI_RESET}"

    def log(self, parts: List[Union[str, Tuple[str, Optional[StyleCode]]]], *,
            is_new_line: bool,
            is_reset: bool):
        tuple_parts = []
        for p in parts:
            if type(p) == str:
                tuple_parts.append((p, None))
            else:
                tuple_parts.append(p)
        coded = [self.__ansi_code(text, color) for text, color in tuple_parts]

        if is_new_line:
            end_char = '\n'
        else:
            end_char = ''

        text = "".join(coded)

        if is_reset:
            print("\r" + text, end=end_char, flush=True)
        else:
            print(text, end=end_char, flush=True)



from typing import Optional, Dict, TYPE_CHECKING

import torch

from labml import tracker
from labml.configs import BaseConfigs

if TYPE_CHECKING:
    from torch.optim.optimizer import Optimizer


def store_l1_l2(name: str, tensor: torch.Tensor):
    if tensor.is_floating_point():
        tracker.add(f"{name}.mean", tensor.mean())
        tracker.add(f"{name}.l1", tensor.abs().mean())
        tracker.add(f"{name}.l2", (tensor ** 2).mean().sqrt())


def store_model_indicators(model: torch.nn.Module, *, model_name: str = "model"):
    for name, param in model.named_parameters():
        if param.requires_grad:
            with torch.no_grad():
                store_l1_l2(f"param.{model_name}.{name}", param)
                if param.grad is not None:
                    store_l1_l2(f"grad.{model_name}.{name}", param.grad)


def store_optimizer_indicators(optimizer: 'Optimizer', *,
                               models: Optional[Dict[str, torch.nn.Module]] = None,
                               optimizer_name: str = "optimizer"):
    if models is None:
        models = {}
    names = {}
    for model_name, model in models.items():
        for name, p in model.named_parameters():
            names[p] = f'{model_name}.{name}'

    unknown = 0
    for group in optimizer.param_groups:
        for p in group['params']:
            if p.grad is None:
                continue
            state = optimizer.state[p]
            if len(state) == 0:
                continue

            name = names.get(p, None)
            if name is None:
                name = f'unknown.{unknown}'
                unknown += 1

            for k, v in state.items():
                if isinstance(v, float) or isinstance(v, int):
                    tracker.add(f'optim.{optimizer_name}.{name}.{k}', v)
                if isinstance(v, torch.Tensor):
                    store_l1_l2(f'optim.{optimizer_name}.{name}.{k}', v)


def get_modules(configs: BaseConfigs):
    keys = dir(configs)

    modules = {}
    for k in keys:
        type_ = configs._get_type(k)
        try:
            if issubclass(type_, torch.nn.Module):
                modules[k] = getattr(configs, k)
        except TypeError:
            pass

    return modules


def get_device(module: torch.nn.Module):
    params = module.parameters()
    try:
        sample_param = next(params)
        return sample_param.device
    except StopIteration:
        raise RuntimeError(f"Unable to determine"
                           f" device of {module.__class__.__name__}") from None



from abc import ABC
from collections import OrderedDict
from typing import Dict, Optional, Any
from uuid import uuid1

import numpy as np

from labml import logger
from labml.internal.util.values import to_numpy
from labml.logger import Text as TextStyle
from . import Indicator

try:
    import matplotlib.pyplot as plt
except (ImportError, ModuleNotFoundError):
    plt = None

try:
    import torch
except ImportError:
    torch = None


class Artifact(Indicator, ABC):
    def __init__(self, *, name: str, is_print: bool):
        super().__init__(name=name, is_print=is_print)

    def _collect_value(self, key: str, value):
        raise NotImplementedError()

    def get_print_length(self) -> Optional[int]:
        raise NotImplementedError()

    @property
    def is_indexed(self) -> bool:
        return False

    def get_string(self, key: str, others: Dict[str, 'Artifact']) -> Optional[str]:
        return None

    def print_all(self):
        pass

    def keys(self):
        raise NotImplementedError()

    def get_value(self, key: str):
        return None

    def collect_value(self, value):
        if type(value) == tuple and len(value) == 2:
            key = value[0]
            value = value[1]
        else:
            key = uuid1().hex

        self._collect_value(key, value)


class _Collection(Artifact, ABC):
    _last_add_step: int
    _values: Dict[str, Any]

    def __init__(self, name: str, is_print: bool = False, density: Optional[float] = None):
        super().__init__(name=name, is_print=is_print)
        self._values = OrderedDict()
        self._density = density
        self._last_add_step = -1

    def _collect_value(self, key: str, value):
        from labml import tracker
        if self._density is None:
            self._values[key] = value
        else:
            steps = tracker.get_global_step() - self._last_add_step
            steps *= self._density
            if np.random.uniform() < 1 - 0.99 ** steps:
                self._values[key] = value

        self._last_add_step = tracker.get_global_step()

    def clear(self):
        self._values = OrderedDict()

    def is_empty(self) -> bool:
        return len(self._values) == 0

    def keys(self):
        return self._values.keys()

    def get_value(self, key: str):
        return self._values[key]


class Tensor(_Collection):
    def __init__(self, name: str, is_once: bool = False):
        super().__init__(name=name, is_print=False)
        self.is_once = is_once

    def to_dict(self) -> Dict:
        return dict(class_name=self.__class__.__name__,
                    name=self.name,
                    is_print=self.is_print,
                    is_once=self.is_once)

    def get_print_length(self) -> Optional[int]:
        return None

    def copy(self, key: str):
        return Tensor(key, is_once=self.is_once)

    def _collect_value(self, key: str, value):
        self._values[key] = to_numpy(value)

    def equals(self, value: any) -> bool:
        if not isinstance(value, Tensor):
            return False
        return value.name == self.name and value.is_once == self.is_once


class Image(_Collection):
    def get_print_length(self) -> Optional[int]:
        return None

    def print_all(self):
        if plt is None:
            logger.log(('matplotlib', logger.Text.highlight),
                       ' not found. So cannot display images')
        images = [to_numpy(v) for v in self._values.values()]
        images = np.concatenate(images)
        cols = min(3, len(images))
        fig: plt.Figure
        fig, axs = plt.subplots((len(images) + cols - 1) // cols, cols,
                                sharex='all', sharey='all',
                                figsize=(8, 10))
        fig.suptitle(self.name)
        for i, img in enumerate(images):
            if len(images) > 1:
                ax: plt.Axes = axs[i // cols, i % cols]
            else:
                ax = axs
            if img.shape[0] == 1:
                img = img[0, :, :]
            else:
                img = img.transpose(1, 2, 0)
            ax.imshow(img)
        plt.show()

    def copy(self, key: str):
        return Image(key, is_print=self.is_print, density=self._density)


class Text(_Collection):
    def get_print_length(self) -> Optional[int]:
        return None

    def print_all(self):
        logger.log(self.name, TextStyle.heading)
        for t in self._values.values():
            logger.log(t, TextStyle.value)

    def copy(self, key: str):
        return Text(key, is_print=self.is_print)

    def equals(self, value: any) -> bool:
        if not isinstance(value, Text):
            return False
        return value.name == self.name and value.is_print == self.is_print


class IndexedText(_Collection):
    def __init__(self, name: str, title: Optional[str] = None, is_print=False):
        super().__init__(name=name, is_print=is_print)
        self.title = title

    @_Collection.is_indexed.getter
    def is_indexed(self) -> bool:
        return True

    def get_print_length(self) -> Optional[int]:
        return max((len(v) for v in self._values.values()))

    def get_string(self, key: str, others: Dict[str, Artifact]) -> Optional[str]:
        return self._values[key]

    def copy(self, key: str):
        return IndexedText(key, title=self.title, is_print=self.is_print)

    def equals(self, value: any) -> bool:
        if not super().equals(value):
            return False
        return value.title == self.title



import pathlib
from typing import Dict

import numpy as np
import torch.nn

from . import ModelSaver, experiment_singleton


class NumpyModelSaver(ModelSaver):
    """
    Deprecated: left for backward compatibility only
    Remove around August 2020
    """

    def __init__(self, name: str, model: torch.nn.Module):
        self.name = name
        self.model = model

    def save(self, checkpoint_path: pathlib.Path) -> any:
        state: Dict[str, torch.Tensor] = self.model.state_dict()
        files = {}
        for key, tensor in state.items():
            if key == "_metadata":
                continue

            file_name = f"{self.name}_{key}.npy"
            files[key] = file_name

            np.save(str(checkpoint_path / file_name), tensor.cpu().numpy())

        return files

    def load(self, checkpoint_path: pathlib.Path, info: any):
        state: Dict[str, torch.Tensor] = self.model.state_dict()
        for key, tensor in state.items():
            file_name = info[key]
            saved = np.load(str(checkpoint_path / file_name))
            saved = torch.from_numpy(saved).to(tensor.device)
            state[key] = saved

        self.model.load_state_dict(state)


class PyTorchModelSaver(ModelSaver):
    def __init__(self, name: str, model: torch.nn.Module):
        self.name = name
        self.model = model

    def save(self, checkpoint_path: pathlib.Path) -> any:
        state = self.model.state_dict()
        file_name = f"{self.name}.pth"
        torch.save(state, str(checkpoint_path / file_name))
        return file_name

    def load(self, checkpoint_path: pathlib.Path, info: any):
        file_name: str = info
        try:
            sample_param = next(self.model.parameters())
            device = sample_param.device
        except StopIteration:
            device = torch.device('cpu')

        state = torch.load(str(checkpoint_path / file_name), map_location=device)

        self.model.load_state_dict(state)


def add_models(models: Dict[str, torch.nn.Module]):
    exp = experiment_singleton()
    savers = {name: PyTorchModelSaver(name, model) for name, model in models.items()}

    exp.checkpoint_saver.add_savers(savers)



"""
labml_remote setup
labml_remote rsync
labml_remote update

Test:

source ~/miniconda/etc/profile.d/conda.sh
conda activate labml_sample_env
cd labml_sample/
PYTHONPATH="${PYTHONPATH}:$(pwd):$(pwd)/src" cmd

export GLOO_SOCKET_IFNAME=enp1s0
export NCCL_SOCKET_IFNAME=enp1s0

PYTHONPATH="${PYTHONPATH}:$(pwd):$(pwd)/src" python labml_samples/pytorch/ddp/launch.py --nproc_per_node=2 --nnodes=2 --node_rank=0 --master_addr="104.171.200.232" --master_port=1234 labml_samples/pytorch/ddp/mnist.py
PYTHONPATH="${PYTHONPATH}:$(pwd):$(pwd)/src" python labml_samples/pytorch/ddp/launch.py --nproc_per_node=2 --nnodes=2 --node_rank=1 --master_addr="104.171.200.232" --master_port=1234 labml_samples/pytorch/ddp/mnist.py


RUN_UUID=fba9eb202cb211eba5beacde48001122 PYTHONPATH="${PYTHONPATH}:$(pwd):$(pwd)/src" python -m torch.distributed.launch --nproc_per_node=2 --nnodes=2 --node_rank=0 --master_addr=104.171.200.181 --master_port=1234 labml_samples/pytorch/ddp/mnist.py
RUN_UUID=fba9eb202cb211eba5beacde48001122 PYTHONPATH="${PYTHONPATH}:$(pwd):$(pwd)/src" python -m torch.distributed.launch --nproc_per_node=2 --nnodes=2 --node_rank=1 --master_addr=104.171.200.181 --master_port=1234 labml_samples/pytorch/ddp/mnist.py

RUN_UUID=fba9eb202cb211eba5beacde48001122 PYTHONPATH="${PYTHONPATH}:$(pwd):$(pwd)/src" python -m torch.distributed.launch --nproc_per_node=2 labml_samples/pytorch/ddp/mnist.py
"""

import time

from labml import experiment
from labml_remote.cli import run_command_async

# run_command_async(0, 'python labml_samples/pytorch/ddp/launch.py --nproc_per_node=2 labml_samples/pytorch/ddp/mnist.py'.split(' '))

run_uuid = experiment.generate_uuid()
for i in [0, 1]:
    # python labml_samples/pytorch/ddp/launch.py --nproc_per_node=2 --nnodes=2 --node_rank=0 --master_addr=104.171.200.232 --master_port=1234 labml_samples/pytorch/ddp/mnist.py
    # run_command_async(i, [f'python labml_samples/pytorch/ddp/launch.py --nproc_per_node=2 --nnodes=2 '
    #                       f'--node_rank={i} --master_addr=104.171.200.181 --master_port=1234 '
    #                       f'labml_samples/pytorch/ddp/mnist.py'], {'GLOO_SOCKET_IFNAME': 'enp1s0'})
    run_command_async(i,
                      [f'python -m torch.distributed.launch --nproc_per_node=2 --nnodes=2 '
                       f'--node_rank={i} '
                       f'--master_addr=104.171.200.181 --master_port=1234 '
                       f'labml_samples/pytorch/ddp/mnist.py'],
                      {'GLOO_SOCKET_IFNAME': 'enp1s0',
                       'RUN_UUID': run_uuid})
    time.sleep(5)



"""
python labml_samples/pytorch/ddp/launch.py --nproc_per_node=3 labml_samples/pytorch/ddp/mnist.py
"""
import os
import subprocess
import sys

from labml.logger import Text

from labml import experiment, logger


def main():
    if 'RUN_UUID' not in os.environ:
        os.environ['RUN_UUID'] = experiment.generate_uuid()

    logger.log(str(sys.argv), Text.danger)
    cmd = [sys.executable, '-u', '-m', 'torch.distributed.launch', *sys.argv[1:]]
    # print(cmd)
    try:
        process = subprocess.Popen(cmd, env=os.environ)
        # print('wait')
        process.wait()
    except Exception as e:
        logger.log('Error starting launcher', Text.danger)
        raise e

    if process.returncode != 0:
        logger.log('Launcher failed', Text.danger)
        raise subprocess.CalledProcessError(returncode=process.returncode, cmd=cmd)


if __name__ == '__main__':
    main()
